/* $Header: /home/dl2/nygren/cvs/pan/src/apps/flood/floodapp.c,v 1.2 1998/02/18 20:00:58 nygren Exp $
 *
 * FILE:    floodapp.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Trivial "flood" application.  Uses floodcapsule and puts 
 * a 4-byte sequence number and a time-stamp into the payload.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <getopt.h>

#include <applib/panapplib.h>
#include <codeobj/portmap/portmap.h>
#include "floodcap_ser_app.h"

#define fail_if(pred,pusage,fmt,arg...) \
   do { if ((pred)) { \
              if(!(pusage)) fprintf(stderr, "%s:%d: ",__FILE__,__LINE__);\
	      fprintf(stderr, fmt, ##arg);\
              if (pusage) print_usage(); \
	      exit(-1); } } while(0)

#define print_verbose(fmt,arg...) \
   if (opt_verbose) \
        fprintf(stderr, fmt,##arg)

#define print_info(fmt,arg...) \
   if (!opt_quiet) \
        printf(fmt, ##arg)


/* Arguments and options */
int         opt_quiet=0;	/* Only output final results */
int         opt_incheader=0;	/* Include headers in bandwidth calculation */
int         opt_usleep=0;	/* Time to sleep between capsules */
int         opt_verbose=0;	/* Be very verbose in printing stuff */
int         opt_count=1;	/* Number of packets to send */
int         opt_multi=0;	/* Whether to send multiple capsules at once */
int         opt_updateperiod=-1;/* How often to we want updates? */
int         opt_capsize=56;	/* The amount of data to add after header */
char       *opt_localnode=NULL;	/* Local user-space node to connect to */
char       *opt_destaddr=NULL; /* Destination address */
int         opt_headersize=0;	/* Header size to add into bandwidth calc */
char       *progname;

/* global variables */
panNodeAddr localaddr, destaddr;
panNodeLink    *pnl;

void flood_send_capsule(panSwseg *capsws, palFloodCapsule *floodcap, int seqno);
int  flood_receive_capsule(void);


void 
print_usage(void)
{
  printf("Usage: %s [-hqvSi] -c <count> -p <period> -m <multi>\n", progname);
  printf("                      -s <size> -z <usleep> -L <localnode>\n");
  printf("  -q              quiet output\n");
  printf("  -v              verbose output\n");
  printf("  -i              include headers in bandwidth calculation\n");
  printf("  -c <count>      number to send\n");
  printf("  -m <multi>      number to send from each inserted capsule\n");
  printf("  -z <usleep>     time to sleep between capsules\n");
  printf("  -p <period>     how often do we want updates?"
	 "  (defaults to <count>)\n"); 
  printf("  -s <size>       total size of capsule (header+payload) not\n");
  printf("                  including underlying headers.\n");
  printf("  -L <localnode>  local user-space node to connect to\n");
}


void
main(int argc, char **argv)
{
  panAppCodeobj  *flood_co;
  panCodeobjName  flood_coname;
  panSwseg       *capsws;
  panTimestamp    rtt, maxrtt={0,0}, minrtt={100000000,0};
  panTimestamp    starttime, endtime, runtime;
  unsigned int    usecaccum=0;
  int32_t         portnum, retval;
  palFloodCapsule  floodcap; 

  int             curopt;	 /* For getopt */
  int             intransit=0; 	 /* Number of capsules in transit */
  int             numreceived=0; /* Number of capsules received */
  int             numsent=0;     /* Number of capsules sent */
  int             seqno=1;       /* Sequence number */
  double          kBps;

  progname=argv[0];

  while (EOF != (curopt = getopt(argc, argv, "ihqvc:p:z:s:L:m:")))
    {
      switch (curopt) 
	{
	case 'h':
	case '?':
	  print_usage(); exit(0);
	  break;
	case 'q':
	  opt_quiet=1;
	  break;
	case 'v':
	  opt_verbose=1;
	  break;
	case 'i':
	  opt_incheader=1;
	  break;
	case 'm':
	  opt_multi=atoi(optarg);
	  break;
	case 'c':
	  opt_count=atoi(optarg);
	  break;
	case 'p':
	  opt_updateperiod=atoi(optarg);
	  break;
	case 's':
	  opt_capsize=atoi(optarg);
	  break;
	case 'z':
	  opt_usleep=atoi(optarg);
	  break;
	case 'L':
	  opt_localnode=optarg;
	  break;
	default:
	  print_usage(); exit(-1);
	}
    }

  if (opt_updateperiod==-1) opt_updateperiod = opt_count;
  if (opt_incheader) {
    opt_headersize = opt_localnode?28:20;
  }
  if (!opt_localnode) opt_localnode="0";

  fail_if((optind != argc-1), 1, "no destination provided\n");
  opt_destaddr = argv[optind];

  print_verbose("Connecting to the node <%s>...\n", opt_localnode);
  pnl = panNodeLink_create(opt_localnode);
  fail_if(!pnl,0, "Couldn't create nodelink!\n");
  panNodeLink_getLocalAddr(&localaddr, pnl);
  print_verbose("Yay!  Connected to local node %s\n",
		panNodeAddr_stringify(NULL,0,&localaddr));

  fail_if(!(panNodeAddr_parse(&destaddr, opt_destaddr)),
	  0,"Couldn't parse dest address!\n");

  portnum = 0;			/* Don't care which... */
  retval = palPortmap_bind(pnl, &portnum, 1/*insertco*/);
  fail_if((retval != 0), 0,"Couldn't bind to port (retval %d)!\n",retval);
  print_verbose("Bound to port number %d\n", portnum);

  flood_co = panAppCodeobj_load("floodcodeobj");
  fail_if(!flood_co,0,"Couldn't load codeobject!\n");

  fail_if(   !panAppCodeobj_getCodeobjName(&flood_coname, flood_co),
	  0, "Couldn't get coname!\n");

  fail_if(   !panAppCodeobj_insertIntoNode(pnl, flood_co, 1),
	  0, "Couldn't insert codeobject!\n");

  if (opt_capsize<palFloodCapsule_SIZE) 
    opt_capsize=palFloodCapsule_SIZE;

  /* Build up the base capsule header than will be used for all capsules */
  capsws = panSwseg_create_byMalloc(opt_capsize);

  panCodeobjName_dup(&(floodcap.coname), &flood_coname);
  panNodeAddr_dup(&(floodcap.cohomeaddr), &localaddr);
  panNodeAddr_dup(&(floodcap.sourceaddr), &localaddr);
  panNodeAddr_dup(&(floodcap.destaddr), &destaddr);
  floodcap.hopcount    = 0;
  floodcap.portnum     = portnum;
  floodcap.capcount    = opt_updateperiod;

  flood_send_capsule(capsws, &floodcap, palFloodCapsule_TYPE_START);
  
  while (numsent < opt_count-1)
    {
      if (opt_multi)
	{
	  floodcap.capcount = opt_multi;
	  flood_send_capsule(capsws, &floodcap, palFloodCapsule_TYPE_MULTI);
	  intransit+=opt_multi;
	  numsent+=opt_multi;
	}
      else
	{
	  flood_send_capsule(capsws, &floodcap, palFloodCapsule_TYPE_DATA);
	  intransit++;
	  numsent++;
	}
      if (opt_usleep) usleep(opt_usleep);
    }

  flood_receive_capsule();

  flood_send_capsule(capsws, &floodcap, palFloodCapsule_TYPE_END);


#if 0
  /* Print summary */
  printf("--- flood statistics ---\n");
  printf("%d packets transmitted, %d packets received, %1.0f%% packet loss\n",
	 numsent, numreceived, 
	 100.0*((double)(numsent-numreceived))/(double)numsent);
  printf("round-trip min/avg/max = %s / %d / ",
	 panTimestamp_stringify(NULL, 0, &minrtt, 0),
	 usecaccum / numreceived);
  printf("%s usecs\n",
	 panTimestamp_stringify(NULL, 0, &maxrtt, 0));	 
  panTimestamp_subtract(&runtime, &endtime, &starttime);
  kBps = ((double)(numreceived*opt_capsize)
    / ((double)runtime.sec)+(((double)runtime.usec)/1e6)) / 1000.0;
  printf("bytes = %d,  runtime = %s secs,   bandwidth = %1.2f KBps = %1.2f Kbps\n", 
	 (numreceived*opt_capsize),
	 panTimestamp_stringify(NULL, 0, &runtime, 
				panTimestamp_STRINGIFY_FMT_INSECS),
	 kBps, 8.0*kBps);
#endif

  print_verbose("Doing a portmapper release...\n");

  palPortmap_release(pnl, portnum);

  print_verbose("Shutting down...\n");

  panNodeLink_disconnect(pnl);

  print_verbose("Destroying...\n");
  panNodeLink_destroy(pnl);
  
}


void 
flood_send_capsule(panSwseg *capsws, palFloodCapsule *floodcap, int type)
{
  panStream       _ps, *ps=&_ps;

  floodcap->type      = type;  
  panStream_init(ps, capsws, 0);
  palFloodCapsule_write(ps, floodcap);
  panStream_init(ps, capsws, 0);
  panNodeLink_send(pnl, ps, opt_capsize);  
}



int
flood_receive_capsule(void)
{
  panStream      _ps, *ps=&_ps;
  panSwseg      *returnedcapsws;
  int            retval;
  palFloodCapsule floodreply;
  panTimestamp   curtime;
  double         kBps;

  print_verbose("flood_receive_capsule: Waiting to get reply back...\n"); 
  returnedcapsws = panNodeLink_retr(pnl, 1);
  print_verbose("flood_receive_capsule: Got reply back!\n");
  panStream_init(ps, returnedcapsws, 0);

  if ((retval=palFloodCapsule_read(&floodreply, ps)) < palFloodCapsule_SIZE)
    { printf("ERROR: got short reply to flood header (%d bytes)\n",retval); }

  print_verbose("Got capsule back with type %d, hop count %d, portnum %d, "
		"capcount %d, bytecount %d, secs %d, usecs %d\n",
		floodreply.type, floodreply.hopcount, 
		floodreply.portnum, floodreply.capcount, floodreply.bytecount,
		floodreply.duration.sec, floodreply.duration.usec);
  if (opt_incheader) 
    {
      floodreply.bytecount += opt_headersize * floodreply.capcount;
    }
  kBps = ((double)(floodreply.bytecount)
    / (((double)floodreply.duration.sec)+(((double)floodreply.duration.usec)/1e6))) / 1000.0;

  print_info("%d bytes from %s:  duration=%s usecs for %d capsules and %d bytes over %d hops   (%0.2f KBps = %0.2f Kbps)\n",
	     returnedcapsws->length, 
	     panNodeAddr_stringify(NULL,0, &floodreply.destaddr),
	     panTimestamp_stringify(NULL, 0, &floodreply.duration, 0),
	     floodreply.capcount, floodreply.bytecount, floodreply.hopcount,
	     kBps, kBps*8.0);

  panSwseg_unref(returnedcapsws);
  return 1;
}
