/* $Header: /home/dl2/nygren/cvs/pan/src/apps/oldping/pingapp.c,v 1.2 1997/12/10 06:58:17 nygren Exp $
 *
 * FILE:    pingapp.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Trivial "ping" application.  Uses pingcapsule and puts 
 * a 4-byte sequence number and a time-stamp into the payload.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>

#include <applib/panapplib.h>
#include <codeobj/portmap/portmap.h>

void
main(int argc, char **argv)
{
  panNodeLink    *pnl;
  panAppCodeobj  *ping_co;
  panCodeobjName  ping_coname;
  panSwseg       *capsws, *returnedcap;
  panStream       _ps, *ps=&_ps;
  panNodeAddr     localaddr, destaddr;
  struct timeval  curtime;
  int seqno=7;
  int32_t         portnum, retval;

  if (argc != 3) 
    { printf("Usage: %s localport destaddr\n", argv[0]);  sleep(3); exit(-1);  }
  if (!argv[1])
    { printf("No node name given\n");  sleep(3); exit(-1);  }
  if (!argv[2])
    { printf("No destination address given\n");  sleep(3); exit(-1);  }

  printf("Connecting to the node <%s>...\n", argv[1]);
  pnl = panNodeLink_create(argv[1]);
  if (!pnl)
    { printf("Couldn't create nodelink!\n"); sleep(3); exit(-1); }
  printf("Yay!  Connected!\n");

  if (!panNodeAddr_parse(&destaddr, argv[2]))
    { printf("Couldn't parse dest address!\n"); sleep(3); exit(-1); }

  portnum = 0;			/* Don't care which... */
  retval = palPortmap_bind(pnl, &portnum, 1/*insertco*/);
  if (retval != 0)
    { printf("Couldn't bind to port (retval %d)!\n",retval);
    sleep(3); exit(-1); }
  printf("Bound to port number %d\n", portnum);

  if (!(ping_co = panAppCodeobj_load("pingcapsule")))
    { printf("Couldn't load codeobject!\n"); sleep(3); exit(-1); }
  if (!panAppCodeobj_getCodeobjName(&ping_coname, ping_co))
    { printf("Couldn't get coname!\n"); sleep(3); exit(-1); }

  if (!panAppCodeobj_insertIntoNode(pnl, ping_co, 1))
    { printf("Couldn't insert codeobject!\n"); sleep(3); exit(-1); }

#define PINGCAPSIZE (panCodeobjName_SIZE + 3*panNodeAddr_SIZE + 4*6)

  capsws = panSwseg_create_byMalloc(PINGCAPSIZE);

  panNodeLink_getLocalAddr(&localaddr, pnl);
  
  panStream_init(ps, capsws, 0);
  panCodeobjName_write(ps, &ping_coname);
  panNodeAddr_write(ps, &localaddr);
  panNodeAddr_write(ps, &localaddr);
  panNodeAddr_write(ps, &destaddr);
  panStream_write_net_int32(ps, 1);
  panStream_write_net_int32(ps, 0);
  panStream_write_net_int32(ps, portnum);
  panStream_write_net_int32(ps, seqno);
  gettimeofday(&curtime, NULL);
  panStream_write_net_uint32(ps, curtime.tv_sec);
  panStream_write_net_uint32(ps, curtime.tv_usec);

  panStream_init(ps, capsws, 0);
  panNodeLink_send(pnl, ps, PINGCAPSIZE);

  returnedcap = panNodeLink_retr(pnl, 1);
  panStream_init(ps, returnedcap, panCodeobjName_SIZE + 3*panNodeAddr_SIZE);
  { 
    int32_t status, hopcount, seqno, portnumreply;
    u_int32_t startsecs, startusecs;
    status = panStream_read_net_int32(ps);
    hopcount = panStream_read_net_int32(ps);
    portnumreply = panStream_read_net_int32(ps);
    seqno = panStream_read_net_int32(ps);
    startsecs = panStream_read_net_uint32(ps);
    startusecs = panStream_read_net_uint32(ps);

    gettimeofday(&curtime, NULL);
    printf("Got capsule back with status %d, hop count %d, portnum %d, "
	   "seq %d, payload 0x%x 0x%x\n", status, hopcount, portnumreply,
	   seqno, startsecs, startusecs);
    printf("Round-trip time = %d usecs\n", (curtime.tv_usec-startusecs)
	   + 1000000*(curtime.tv_sec-startsecs));
  }

  palPortmap_release(pnl, portnum);

  panNodeLink_disconnect(pnl);
  panNodeLink_destroy(pnl);
  
}
