/* $Header: /home/dl2/nygren/cvs/pan/src/apps/oldping/pingcapsule.c,v 1.1 1997/12/01 22:18:03 nygren Exp $
 *
 * FILE:    pingcapsule.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Very rudimentry ping capsule for experimental use.
 *
 * Capsule contains:
 * 
 *    SIZE        DESCRIPTION
 *    ----        -----------
 *    24 bytes    CodeobjName --     Name of this code object
 *    16 bytes    CodeobjHomeAddr -- Home for this code object
 *
 *    16 bytes    SourceAddr         The source of the ping
 *    16 bytes    DestAddr           The destination of the ping
 *     4 bytes    Status             What this capsule is doing:
 *                                       1 == just inserted 
 *                                       2 == going to dest   
 *                                       3 == going back to source
 *     4 bytes    HopCount           Number of nodes traversed
 *     4 bytes    PortNumber         Port number of application to deliver to
 *     N bytes    ArbitraryPayload   Arbitrary filler (returned to app)
 *
 * NOTE: Doesn't do much error-checking yet.
 *
 */

#include <common/pan-copyright.h>
#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pninetwork.h>
#include <pni/pniapplink.h>
#include <pni/pnicodeobj.h>
#include <codeobj/portmap/portmap.h>



void accept(pniStream *capstart, int caplength)
{
  pniNodeAddr curaddr, dstaddr, srcaddr;
  int32_t     capstatus, hopcount, portnum; 
  pniStream   _ps, _statpos, *ps=&_ps, *statpos=&_statpos;
  pniAppLink  *theapp;

  pniStream_dup(ps, capstart);

  pniStream_seekFwd(ps, pniCapsuleHeader_SIZE);
  
  pniNodeAddr_read(&srcaddr, ps);
  pniNodeAddr_read(&dstaddr, ps);

  pniStream_dup(statpos, ps);
  capstatus = pniStream_read_net_int32(ps);
  hopcount  = pniStream_read_net_int32(ps);
  hopcount++;
  portnum   = pniStream_read_net_int32(ps);
  printf("---> At node with hopcount %d, portnum %d\n", hopcount, portnum);

  pniNodeAddr_getLocalAddr(&curaddr);

  if (capstatus == 1)		/* Just inserted */
    {
      printf("---> STARTING PING!!!\n");

      capstatus = 2;
    }

  if (capstatus == 2)	/* Heading towards destination */
    {
      if (0 == pniNodeAddr_compare(&dstaddr, &curaddr))
	{
	  printf("---> AT DESTINATION!!!\n");
	  capstatus = 3;
	}
      else 
	{
	  printf("---> HEADING TO DESTINATION!!!\n");
	  pniStream_write_net_int32(statpos, capstatus);
	  pniStream_write_net_int32(statpos, hopcount);
	  pniNetwork_sendCapsule(&dstaddr, capstart, caplength);
	  return;
	}
    }

  if (capstatus == 3)	/* Heading towards source */
    {
      if (0 == pniNodeAddr_compare(&srcaddr, &curaddr))
	{
	  printf("---> BACK AT SOURCE!!!\n");
	  pniStream_write_net_int32(statpos, capstatus);
	  pniStream_write_net_int32(statpos, hopcount);

	  pniStream_dup(ps, capstart);

	  theapp = pcoPortmap_mapPortnumToApplink(portnum);
	  if (!theapp)
	    { printf("---> ERROR: couldn't find port %d\n", portnum); return; }
	  pniAppLink_deliverData(theapp, capstart, caplength);

	  { 
	    int32_t status, hopcount, seqno;
	    u_int32_t payload1, payload2;


	    pniStream_seekFwd(ps, panCodeobjName_SIZE + 3*panNodeAddr_SIZE);
	    status = pniStream_read_net_int32(ps);
	    hopcount = pniStream_read_net_int32(ps);
	    seqno = pniStream_read_net_int32(ps);
	    payload1 = pniStream_read_net_uint32(ps);
	    payload2 = pniStream_read_net_uint32(ps);
	    
	    printf("---> Capsule has status %d, hop count %d, seq %d, "
		   "payload 0x%x 0x%x\n", status, hopcount, seqno, payload1, payload2);
	  }


	  return;
	}
      else 
	{
	  printf("---> HEADING HOME!!!\n");
	  pniStream_write_net_int32(statpos, capstatus);
	  pniStream_write_net_int32(statpos, hopcount);
	  pniNetwork_sendCapsule(&srcaddr, capstart, caplength);
	  return;
	}
    }
}

