#!/usr/bin/perl

@packet_sizes = ( 128, 256, 512, 1024, 1500, 1504, 2048, 4096, 8192);

############################## headersize, localport, destination
%configmap = ( 
	      "kernelactive"  => [ 20, "",     "10.1.0.8" ],
	      "kernelactivecpy"  => [ 20, "",     "10.1.0.8" ],
	      "kernelpassive" => [ 20, "",     "10.1.0.8" ],
	      "kernelnone"    => [ 20, "",     "10.0.0.8" ],
	      "useractive"    => [ 28, "3250", "10.1.0.8:3250" ],
	      "userpassive"   => [ 28, "3250", "10.1.0.8:3250" ],
	      "usernone"      => [ 28, "3250", "10.0.0.8:3250" ],
	      "usercforw"     => [ 28, "3250", "10.1.0.8:3250" ],
	      "testactive"    => [ 28, "3256", "mindwipe:9764" ]
	      );

sub do_ping
{
    my ($count, $size, $dest, $local) = @_;
    my ($localopt, $rttmin, $rttavg, $rttmax, $totalsize, $realsize, $pvers) 
	= ("", 0, 0, 0, 0, 0, ".kernel");

    if ($local) {
	$localopt = "-L $local";
	$pvers = ".userspace";
    }

    open PINGOUT, "./pingapp$pvers -q -c $count -s $size $localopt $dest |" or die "couldn't start up ping";
    
    while (<PINGOUT>) 
    {
	if (/^round-trip min\/avg\/max = ([0-9]+) \/ ([0-9]+) \/ ([0-9]+)/) {
	    $rttmin = $1; $rttavg = $2; $rttmax = $3;
	}
	if (/^bytes = ([0-9]+),/) {
	    $totalsize = $1; $realsize = $totalsize / $count;
	}
    }

    close PINGOUT;
    #print "results: [$size,$realsize] $rttmin $rttavg $rttmax  ($localopt)\n";
    return ($realsize, $count, $rttmin, $rttavg, $rttmax);
}


#do_ping(5000, 1000, "mindwipe:9764", 3256);

$nodetype = $ARGV[0];
$nettype  = $ARGV[1];
$runcount = $ARGV[2];
$testtype = ($nodetype.$nettype);
(("" eq $nodetype) || ("" eq $nettype)) and die "invalid arguments\n";
if ($runcount eq "") { $runcount = 5000; }

open RESULTS, ">ping.results.$testtype.csv" or die "couldn't open output file";
print "outputting results to ping.results.$testtype.csv\n";

$configmap{$testtype} or die "unknown test type <$testtype>\n";
print "test type = $testtype ($nodetype+$nettype)\n";

($headersize, $localport, $destination) = @{$configmap{$testtype}};

$header = sprintf "%8s, %8s, %6s, %6s, %6s, %6s, %6s, %6s\n",
    "nodetype", "nettype", "runcnt", "paksiz", "capsiz", 
    "rttmin", "rttavg", "rttmax";
print RESULTS $header;
print $header;

# Prime the link first...
foreach $currun (1..3) {
    ($retcapsize, $ret_count, $rttmin, $rttavg, $rttmax) = 
	do_ping(100, 0, $destination, $localport);
    $results = sprintf "%8s, %8s, %6d, %6d, %6d, %6d, %6d, %6d\n",
    $nodetype, $nettype, $ret_count, $retcapsize+headersize, 
    $retcapsize, $rttmin, $rttavg, $rttmax;
    print RESULTS $results;
    print $results;
    sleep(2);
}

foreach $curpaksize (@packet_sizes)
{
    if ($curpaksize < 4000) {
	print "skipping $curpaksize...\n";
	next;
    }
    foreach $currun (1..3) 
    {
	($retcapsize, $ret_count, $rttmin, $rttavg, $rttmax) = 
	    do_ping($runcount, $curpaksize-$headersize, 
		    $destination, $localport);
	$results = sprintf "%8s, %8s, %6d, %6d, %6d, %6d, %6d, %6d\n",
	$nodetype, $nettype, $ret_count, $retcapsize+$headersize, 
	$retcapsize, $rttmin, $rttavg, $rttmax;
	print $results;
	print RESULTS $results;
	sleep(2);	
    }
}

close RESULTS;
