/* $Header: /home/dl2/nygren/cvs/pan/src/apps/ping/pingcapsule.c,v 1.3 1998/01/20 23:43:04 nygren Exp $
 *
 * FILE:    pingcapsule.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Very rudimentry ping capsule for experimental use.
 * See pingcap.psg for the specification of the capsule contents.
 *
 * NOTE: Doesn't do much error-checking yet.
 *
 */

#include <common/pan-copyright.h>
#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pninetwork.h>
#include <pni/pniapplink.h>
#include <pni/pnicodeobj.h>
#include <pni/pnimisc.h>
#include <codeobj/portmap/portmap.h>
#include "pingcap_ser_co.h"


void accept(pniStream *capstart, int caplength)
{
  pniStream   _ps, *ps=&_ps;
  pniNodeAddr curaddr;
  pniAppLink  *theapp;
  pcoPingCapsule pingcap;

  pniStream_dup(ps, capstart);
  pcoPingCapsule_read(&pingcap, ps);
  pniStream_dup(ps, capstart);

  pingcap.hopcount++;
  pniLog_debug("---> At node with hopcount %d, portnum %d\n", 
	 pingcap.hopcount, pingcap.portnum);

  pniNodeAddr_getLocalAddr(&curaddr);

  if (pingcap.status == pcoPingCapsule_STATUS_JUST_INSERTED)
    {
      pniLog_debug("---> STARTING PING!!!\n");
      pingcap.status = pcoPingCapsule_STATUS_GOING_TO_DST;
      if (pingcap.stampatnode)	
	pniTimestamp_getCurrent(&pingcap.senttime);
    }

  if (pingcap.status == pcoPingCapsule_STATUS_GOING_TO_DST)
    {
      if (0 == pniNodeAddr_compare(&pingcap.destaddr, &curaddr))
	{
	  pniLog_debug("---> AT DESTINATION!!!\n");
	  pingcap.status = pcoPingCapsule_STATUS_GOING_TO_SRC;
	}
      else 
	{
	  pniLog_debug("---> HEADING TO DESTINATION!!!\n");
	  pcoPingCapsule_write(ps, &pingcap);
	  pniNetwork_sendCapsule(&pingcap.destaddr, capstart, caplength);
	  return;
	}
    }

  if (pingcap.status == pcoPingCapsule_STATUS_GOING_TO_SRC)
    {
      if (0 == pniNodeAddr_compare(&pingcap.sourceaddr, &curaddr))
	{
	  if (pingcap.stampatnode)	
	    {
	      pniTimestamp curtime;
	      pniTimestamp_getCurrent(&curtime);
	      pniTimestamp_subtract(&pingcap.senttime,
				    &curtime, &pingcap.senttime);
	    }

	  pniLog_debug("---> BACK AT SOURCE!!!\n");
	  pcoPingCapsule_write(ps, &pingcap);

	  theapp = pcoPortmap_mapPortnumToApplink(pingcap.portnum);
	  if (!theapp)
	    { pniLog_debug("---> ERROR: couldn't find port %d\n", pingcap.portnum); 
	    return; }
	  pniAppLink_deliverData(theapp, capstart, caplength);
	  return;
	}
      else 
	{
	  pniLog_debug("---> HEADING HOME!!!\n");
	  pcoPingCapsule_write(ps, &pingcap);
	  pniNetwork_sendCapsule(&pingcap.sourceaddr, capstart, caplength);
	  return;
	}
    }
}

