/* $Header: /home/dl2/nygren/cvs/pan/src/apps/simpleping/pingapp.c,v 1.1 1998/01/20 23:43:16 nygren Exp $
 *
 * FILE:    pingapp.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Trivial "ping" application.  Uses pingcapsule and puts 
 * a 4-byte sequence number and a time-stamp into the payload.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include <applib/panapplib.h>
#include <codeobj/portmap/portmap.h>
#include "pingcap_ser_app.h"

void
main(int argc, char **argv)
{
  panNodeLink    *pnl;
  panAppCodeobj  *ping_co;
  panCodeobjName  ping_coname;
  panSwseg       *capsws, *returnedcapsws;
  panStream       _ps, *ps=&_ps;
  panNodeAddr     localaddr, destaddr;
  panTimestamp    curtime, rtt;
  int seqno=7;
  int32_t         portnum, retval;
  palPingCapsule  pingcap, pingreply;

  if (argc != 3) 
    { printf("Usage: %s localport destaddr\n", argv[0]);  sleep(3); exit(-1);  }
  if (!argv[1])
    { printf("No node name given\n");  sleep(3); exit(-1);  }
  if (!argv[2])
    { printf("No destination address given\n");  sleep(3); exit(-1);  }

  printf("Connecting to the node <%s>...\n", argv[1]);
  pnl = panNodeLink_create(argv[1]);
  if (!pnl)
    { printf("Couldn't create nodelink!\n"); sleep(3); exit(-1); }
  printf("Yay!  Connected!\n");
  printf("Local node address: ");
  panNodeLink_getLocalAddr(&localaddr, pnl);
  panNodeAddr_print(&localaddr); printf("\n");

  if (!panNodeAddr_parse(&destaddr, argv[2]))
    { printf("Couldn't parse dest address!\n"); sleep(3); exit(-1); }

  portnum = 0;			/* Don't care which... */
  retval = palPortmap_bind(pnl, &portnum, 1/*insertco*/);
  if (retval != 0)
    { printf("Couldn't bind to port (retval %d)!\n",retval);
    sleep(3); exit(-1); }
  printf("Bound to port number %d\n", portnum);

  if (!(ping_co = panAppCodeobj_load("pingcapsule")))
    { printf("Couldn't load codeobject!\n"); sleep(3); exit(-1); }
  if (!panAppCodeobj_getCodeobjName(&ping_coname, ping_co))
    { printf("Couldn't get coname!\n"); sleep(3); exit(-1); }

  if (!panAppCodeobj_insertIntoNode(pnl, ping_co, 1))
    { printf("Couldn't insert codeobject!\n"); sleep(3); exit(-1); }

  capsws = panSwseg_create_byMalloc(palPingCapsule_SIZE+1000);

  panNodeLink_getLocalAddr(&localaddr, pnl);

  panCodeobjName_dup(&(pingcap.coname), &ping_coname);
  panNodeAddr_dup(&(pingcap.cohomeaddr), &localaddr);
  panNodeAddr_dup(&(pingcap.sourceaddr), &localaddr);
  panNodeAddr_dup(&(pingcap.destaddr), &destaddr);
  pingcap.status     = palPingCapsule_STATUS_JUST_INSERTED;
  pingcap.hopcount   = 0;
  pingcap.portnum    = portnum;
  pingcap.seqno      = seqno;
  panTimestamp_getCurrent(&pingcap.senttime);

  panStream_init(ps, capsws, 0);
  palPingCapsule_write(ps, &pingcap);
  panStream_init(ps, capsws, 0);
  panNodeLink_send(pnl, ps, palPingCapsule_SIZE+1000);

  panPrint_debug("Waiting to get reply back...\n"); 
  returnedcapsws = panNodeLink_retr(pnl, 1);
  panTimestamp_getCurrent(&curtime);
  panPrint_debug("Got reply back!\n");
  panStream_init(ps, returnedcapsws, 0);
  if (retval=palPingCapsule_read(&pingreply, ps) < palPingCapsule_SIZE+1000)
    { printf("ERROR: got short reply to ping (%d bytes)\n",retval); }

  printf("Got capsule back with status %d, hop count %d, portnum %d, "
	 "seq %d, payload 0x%x 0x%x\n", pingreply.status, pingreply.hopcount, 
	 pingreply.portnum, pingreply.seqno,
	 pingreply.senttime.sec, pingreply.senttime.usec);
  panTimestamp_subtract(&rtt, &curtime, &pingreply.senttime);
#if 0
  printf("Round-trip time = %d usecs\n", (curtime.tv_usec-pingreply.sent_usecs)
	 + 1000000*(curtime.tv_sec-pingreply.sent_secs));
#endif
  printf("Round-trip time = %s usecs\n", 
	 panTimestamp_stringify(NULL, 0, &rtt, 0));


#if 0
  { 
    int32_t status, hopcount, seqno, portnumreply;
    u_int32_t startsecs, startusecs;
    status = panStream_read_net_int32(ps);
    hopcount = panStream_read_net_int32(ps);
    portnumreply = panStream_read_net_int32(ps);
    seqno = panStream_read_net_int32(ps);
    startsecs = panStream_read_net_uint32(ps);
    startusecs = panStream_read_net_uint32(ps);

    gettimeofday(&curtime, NULL);
    printf("Got capsule back with status %d, hop count %d, portnum %d, "
	   "seq %d, payload 0x%x 0x%x\n", status, hopcount, portnumreply,
	   seqno, startsecs, startusecs);
    printf("Round-trip time = %d usecs\n", (curtime.tv_usec-startusecs)
	   + 1000000*(curtime.tv_sec-startsecs));
  }
#endif

  printf("Doing a portmapper release...\n");

  palPortmap_release(pnl, portnum);

#if 0
  printf("Sleeping before shutting down...\n");
  sleep(1);
#endif

  printf("Shutting down...\n");

  panNodeLink_disconnect(pnl);

  printf("Destroying...\n");
  panNodeLink_destroy(pnl);
  
}
