/* $Header: /home/dl2/nygren/cvs/pan/src/codeobj/insertco/insertco.c,v 1.4 1998/01/20 23:43:37 nygren Exp $
 *
 * FILE:    insertco.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Inserts a code object into a node's code cache.
 *
 * Capsule contains:
 * 
 *    SIZE        DESCRIPTION
 *    ----        -----------
 *    24 bytes    ICOCodeobjName --     Name of this dynamicload code object
 *    16 bytes    ICOCodeobjHomeAddr -- Home for this dynamicload code object
 *     4 bytes    PinCOP                1 if code object should be pinned in 
 *                                      code cache, else 0
 *    16 bytes    ICHomeAddr            Home for the inserted code
 *     N bytes    CodeobjToInsert       The code object to insert
 *
 * NOTE: Doesn't do much error-checking yet.
 *
 */

#include <common/pan-copyright.h>
#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pninetwork.h>
#include <pni/pnicodeobj.h>
#include <pni/pnimisc.h>


/* Just prints out a message to indicate that it has been
 * loaded.  Useful since this may be the first code object
 * to get loaded into the system. */
void
init(void)
{
  int x, y=3;
  x = 5+y;
  pniLog_debug("--> Init method of insertco code object called!  5+3=%d=%d\n",
	       x,y+5);
}


/* Accepts and evaluates a dynamicload capsule */
void 
accept(pniStream *capstart, int caplength)
{

  pniStream       _ps, *ps=&_ps;
  int             pinP;
  pniNodeAddr     homeaddr;

  pniStream_dup(ps, capstart);
  /* Seek forward past the capsule header */
  pniStream_seekFwd(ps, pniCapsuleHeader_SIZE);

  /* Read whether to pin the capsule in the code cache */
  pinP = pniStream_read_net_int32(ps);

  /* Where to get dependencies from */
  pniNodeAddr_read(&homeaddr, ps);

  pniLog_debug("insertco: loading codeobj into node\n");
  pniCodeobj_load(ps, &homeaddr);
  
  if (pinP) pniLog_debug("insertco: need to pin, but don't know how to\n");
  
}

