/* $Header: /home/dl2/nygren/cvs/pan/src/codeobj/portmap/pal_portmap.c,v 1.1 1997/12/10 06:58:20 nygren Exp $
 *
 * FILE:    pingapp.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Trivial "ping" application.  Uses pingcapsule and puts 
 * a 4-byte sequence number and a time-stamp into the payload.
 *
 */

#include <applib/panapplib.h>
#include "portmap.h"

static panAppCodeobj  *portmap_co = NULL;
static panCodeobjName  portmap_coname;



/* Binds a nodelink/applink to a portnumber.
 * Should be the first operation performed over the link
 * as it needs to receive a reply from the portmap codeobj.
 * If portnum is 0, binds to an arbitrary port.  Otherwise,
 * tries to bind to portnum.  Portnum is used to return
 * the port actually bound it.  Insertco should be 1 if
 * the portmap code object needs to be inserted into the
 * node, else 0.  Returns a return value (0 on success).
 */
int
palPortmap_bind(panNodeLink *nl, int *portnum, int insertco)
{
  panSwseg *capsws, *replsws;
  panStream       _ps, *ps=&_ps;
  panNodeAddr     localaddr;
  panCodeobjName  replconame;
  int newportnum, retval;

  if (!portmap_co)
    {
      portmap_co = panAppCodeobj_load("portmap");
      if (!portmap_co)
	{ panPrint_warn("palPortmap_bind: couldn't load codeobj\n");
	return pcoPortmap_RVAL_ELINKFAIL; }
      if (!panAppCodeobj_getCodeobjName(&portmap_coname, portmap_co))
	{ panPrint_warn("palPortmap_bind: Couldn't get coname!\n"); 
	return pcoPortmap_RVAL_ELINKFAIL; }
    }

  if (insertco)
    {
      if (!panAppCodeobj_insertIntoNode(nl, portmap_co, 1))
	{ panPrint_warn("palPortmap_bind: couldn't insert codeobject!\n"); 
	return pcoPortmap_RVAL_ELINKFAIL; }
    }

#define PORTMAPCAPSIZE (panCodeobjName_SIZE + panNodeAddr_SIZE + 4*2)
  capsws = panSwseg_create_byMalloc(PORTMAPCAPSIZE);
  if (!capsws) return pcoPortmap_RVAL_EFULL;
  panNodeLink_getLocalAddr(&localaddr, nl);
  
  panStream_init(ps, capsws, 0);
  panCodeobjName_write(ps, &portmap_coname);
  panNodeAddr_write(ps, &localaddr);
  panStream_write_net_int32(ps, pcoPortmap_OP_BIND);
  panStream_write_net_int32(ps, *portnum);

  panStream_init(ps, capsws, 0);
  if (PORTMAPCAPSIZE != panNodeLink_send(nl, ps, PORTMAPCAPSIZE))
    return pcoPortmap_RVAL_ELINKFAIL;
  panSwseg_unref(capsws);
  
  while (replsws = panNodeLink_retr(nl, 1))
    {
      panStream_init(ps, replsws, 0);
      panCodeobjName_read(&replconame, ps, 0);
      if (panCodeobjName_equalP(&replconame, &portmap_coname))
	break;
      else
	panPrint_debug("palPortmap_bind: discarded unknown reply\n"); 
      panSwseg_unref(replsws);	  
    }
  if (!replsws) 
    return pcoPortmap_RVAL_ELINKFAIL;

  panStream_seekFwd(ps, panNodeAddr_SIZE);
  
  retval     = panStream_read_net_int32(ps);
  newportnum = panStream_read_net_int32(ps);

  *portnum = newportnum;

  panSwseg_unref(replsws);

  return retval;
}


/* Deletes a mapping to a port number.  
 * Returns the status of the operation.  
 *
 * *TO*DO* Currently waits for a response and returns
 * it.  Any other data received that's not from the portmap
 * code object will be dropped.  
 */
int
palPortmap_release(panNodeLink *nl, int portnum)
{
  panSwseg *capsws, *replsws;
  panStream       _ps, *ps=&_ps;
  panNodeAddr     localaddr;
  panCodeobjName  replconame;
  int newportnum, retval;

  if (!portmap_co)
    { panPrint_warn("palPortmap_release: called without codeobj loaded\n");
    return pcoPortmap_RVAL_ELINKFAIL; }

  capsws = panSwseg_create_byMalloc(PORTMAPCAPSIZE);
  if (!capsws) return pcoPortmap_RVAL_EFULL;
  panNodeLink_getLocalAddr(&localaddr, nl);
  
  panStream_init(ps, capsws, 0);
  panCodeobjName_write(ps, &portmap_coname);
  panNodeAddr_write(ps, &localaddr);
  panStream_write_net_int32(ps, pcoPortmap_OP_DELETE);
  panStream_write_net_int32(ps, portnum);

  panStream_init(ps, capsws, 0);
  if (PORTMAPCAPSIZE != panNodeLink_send(nl, ps, PORTMAPCAPSIZE))
    return pcoPortmap_RVAL_ELINKFAIL;
  panSwseg_unref(capsws);
  
  while (replsws = panNodeLink_retr(nl, 1))
    {
      panStream_init(ps, replsws, 0);
      panCodeobjName_read(&replconame, ps, 0);
      if (panCodeobjName_equalP(&replconame, &portmap_coname))
	break;
      else
	panPrint_debug("palPortmap_release: discarded unknown reply\n"); 
      panSwseg_unref(replsws);	  
    }
  if (!replsws) 
    return pcoPortmap_RVAL_ELINKFAIL;

  panStream_seekFwd(ps, panNodeAddr_SIZE);
  
  retval     = panStream_read_net_int32(ps);
  newportnum = panStream_read_net_int32(ps);

  panSwseg_unref(replsws);

  return retval;
}

