/* $Header: /home/dl2/nygren/cvs/pan/src/common/nodeaddr.h,v 1.5 1998/01/20 23:44:18 nygren Exp $
 *
 * FILE:    nodeaddr.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 *
 */

#ifndef NODE_NODEADDR_H_INC
#define NODE_NODEADDR_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <common/sysconfig.h>


/* Number of bytes in a PAN Node Address */
#define panNodeAddr_SIZE 16

/* An internal representation for a PAN node address.
 * In capsules, all PAN addresses are 128 bits. */
typedef struct _panNodeAddr {
  u_int8_t data[panNodeAddr_SIZE]; /* The address for a node is 128 bits.
				    * The contents depend on the type of 
				    * node. */
  /* For UNode's: 4-byte IP addr followed by 2-byte portnum then padding. */
  /* For IPv4 kernel nodes: 4-byte IP addr followed by 12 bytes of padding. */
  /* For IPv6 kernel nodes: 16-byte IP addr */
} panNodeAddr;


/* These functions are implemented in machine-specific code in the
 * arch directory... */

extern unsigned
panNodeAddr_read(panNodeAddr *na, panStream *ps);

extern unsigned
panNodeAddr_write(panStream *ps, panNodeAddr *na);

extern void 
panNodeAddr_print(panNodeAddr *na);

extern char *
panNodeAddr_stringify(char *buf, int buflen, panNodeAddr *na);

extern int
panNodeAddr_compare(panNodeAddr *na1, panNodeAddr *na2);

extern void
panNodeAddr_getLocalAddr(panNodeAddr *na);

extern int
panNodeAddr_parse(panNodeAddr *na, char *nodename);


#define panNodeAddr_dup(dst, src) \
        memcpy((void*)(dst), (void*)(src), sizeof(panNodeAddr))


#endif /* NODE_NODEADDR_H_INC */
