/* $Id: sha1.h,v 1.1.1.1 1997/05/14 00:38:32 nygren Exp $ */

#ifndef _SHA1_H_
#define _SHA1_H_

/*
   SHA-1 in C
   By Steve Reid <steve@edmweb.com>
   100% Public Domain

   Test Vectors (from FIPS PUB 180-1)
   "abc"
   A9993E36 4706816A BA3E2571 7850C26C 9CD0D89D
   "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
   84983E44 1C3BD26E BAAE4AA1 F95129E5 E54670F1
   A million repetitions of "a"
   34AA973C D4C4DAA4 F61EEB2B DBAD2731 6534016F
 */

#include "common/sysconfig.h"
#include <sys/types.h>

#define SHA1_SIZE 20
#define SHA1_BITSIZE (SHA1_SIZE*8)

typedef struct {
   u_int32_t state[5];
#ifdef HAVE_U_INT64_T
   u_int64_t count;
#else /* !HAVE_U_INT64_T */
   /* Note, this is non-compliant with SHA-1 standard.*/
   /* Messages limited to 4Gb */
   u_int32_t count;
#endif/* !HAVE_U_INT64_T */
   u_int8_t buffer[64];
} sha1_ctx;

void sha1_init (sha1_ctx *context);
void sha1_update (sha1_ctx *context, const u_char *data, u_int len);
void sha1_final (u_char digest[20], sha1_ctx *context);
void sha1_transform (u_int32_t state[5], const u_int8_t buffer[64]);

#endif /* _SHA1_H_ */
