/* $Header: /home/dl2/nygren/cvs/pan/src/common/streamwordio.h,v 1.2 1997/12/10 06:58:22 nygren Exp $
 *
 * FILE:    swsstream.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for software segment stream data structures and routines.
 * See PAN/doc/software-segments.txt for a detailed description
 * of software segments.  Be sure to read that file before trying to
 * use software segment streams.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>

/*
 * Initializes the panStream object ps to start pos bytes into sws.
 * (if pos==0, starts at start os sws)
 */
void
panStream_init(panStream *ps, panSwseg *sws, unsigned pos)
{
  while ((pos > sws->length) && sws)
    {
      pos -= sws->length;
      sws = sws->next;
    }
  assert(ps&&sws);
  ps->pos = sws->data + pos;
  ps->rem = sws->length - pos;
  ps->cursws = sws;     
}


/*
 * Seeks the current position forward by offset bytes
 */
void
panStream_seekFwd(panStream *ps, unsigned offset)
{
  while (1 /*offset > 0*/) 
    {
      if (ps->rem < offset)
	{
	  if (ps->cursws->next)	/* Skip to next sws */
	    {
	      offset -= ps->rem;
	      ps->cursws = ps->cursws->next;
	      ps->rem = ps->cursws->length;	      
	      ps->pos = ps->cursws->data;
	    }
	  else			/* There is no next segment.  Seek to 
				 * end and return.  */
	    {
	      ps->rem = 0;
	      ps->pos += ps->rem;
	      return;
	    }
	}
      else			/* Seek to somewhere in current swseg */
	{
	  ps->rem -= offset;
	  ps->pos += offset;
	  /* offset = 0; */
	  return;
	}
    }
}


/*
 * Returns the current software segment and stores the position
 * of the stream into pos.
 */
panSwseg *
panStream_getState(panStream *ps, unsigned *curpos)
{
  if (!ps) return NULL;
  *curpos = ps->cursws->length - ps->rem;
  return ps->cursws;
}


/*
 * Copies length bytes from pssrc into psdest and returns
 * the number of bytes copied (which may be less than length
 * if either is shorter than length).  Seeks forward both
 * streams by length.
 */
unsigned 
panStream_copy(panStream *psdest, panStream *pssrc, unsigned length)
{
  unsigned numcopied = 0;

  while (length > 0)
    {
      if ((pssrc->rem == 0 && !pssrc->cursws->next) 
	  || (psdest->rem == 0 && !psdest->cursws->next))
	return numcopied;
      else if ((length <= pssrc->rem) && (length <= psdest->rem))
	{
	  memcpy(psdest->pos, pssrc->pos, length);
	  pssrc->rem  -= length;
	  psdest->rem -= length;
	  pssrc->pos  += length;
	  psdest->pos += length;
	  numcopied   += length;
	  return numcopied;
	}
      else 
	{
	  unsigned tocopy = MIN(pssrc->rem, psdest->rem);
	  tocopy = MIN(tocopy, length);
	  
	  memcpy(psdest->pos, pssrc->pos, tocopy);
	  pssrc->rem  -= tocopy;
	  psdest->rem -= tocopy;
	  pssrc->pos  += tocopy;
	  psdest->pos += tocopy;
	  numcopied   += tocopy;
	  length      -= tocopy;	  
	  if ((pssrc->rem == 0) && pssrc->cursws->next)
	    {
	      pssrc->cursws = pssrc->cursws->next;
	      pssrc->rem    = pssrc->cursws->length;
	      pssrc->pos    = pssrc->cursws->data;
	    }
	  if ((psdest->rem == 0) && psdest->cursws->next)
	    {
	      psdest->cursws = psdest->cursws->next;
	      psdest->rem    = psdest->cursws->length;
	      psdest->pos    = psdest->cursws->data;
	    }
	}
    }
  return numcopied;
}


/*
 * Returns whether a panStream is at the end of the software segment chain.
 * (Returns 1 iff at end, else returns 0).
 */
int
panStream_eofP(panStream *ps)
{
  if (!ps || (ps->rem == 0) || !(ps->cursws))
    return 1;
  else
    return 0;
}


#if 0
u_int32_t 
panStream_read_msbfirst_uint32(panStream *ps)
{
  if (ps->rem >= 4)
    {
      ps->rem -= 4;
      return ntohl(*(((u_int32_t*)(ps->pos))++));
    }
  else if (ps->rem != 0)
    {
      panPrint_error("panStream_read_msbfirst_uint32: unaligned read\n");
      return 0;
    }
  else if (ps->cursws->next)
    {
      ps->cursws = ps->cursws->next;
      ps->rem = ps->cursws->length - 4;
      ps->pos = ps->cursws->data;
      return ntohl(*(((u_int32_t*)(ps->pos))++));
    }
  else
    {
      panPrint_error("panStream_read_msbfirst_uint32: read of empty stream\n");
      return 0;
    }
}
#endif

#ifdef panStream_MAKE_PROTOS
#define __panStream_make_read(FNNAME, BYTES, TYPE, CONVERTFN) \
    extern TYPE FNNAME(panStream *ps);
#define __panStream_make_write(FNNAME, BYTES, TYPE, CONVERTFN) \
    extern void FNNAME(panStream *ps, TYPE x);
#else

#define __panStream_make_read(FNNAME, BYTES, TYPE, CONVERTFN) \
    TYPE FNNAME(panStream *ps) { \
	if (ps->rem >= BYTES) { \
             ps->rem -= BYTES; \
             return CONVERTFN(*(((TYPE*)(ps->pos))++)); \
        } else if (ps->rem != 0) { \
             panPrint_error( #FNNAME ": unaligned read\n"); return 0; \
        } else if (ps->cursws->next) { \
             ps->cursws = ps->cursws->next; \
             ps->rem = ps->cursws->length - BYTES; \
             ps->pos = ps->cursws->data; \
             return CONVERTFN(*(((TYPE*)(ps->pos))++)); \
        } else { \
             panPrint_error( #FNNAME ": read of empty stream\n"); \
             return 0; \
        }  \
     }

#define __panStream_make_write(FNNAME, BYTES, TYPE, CONVERTFN) \
    void FNNAME(panStream *ps, TYPE x) { \
	if (ps->rem >= BYTES) { \
             ps->rem -= BYTES; \
             *(((TYPE*)(ps->pos))++) = CONVERTFN(x); \
        } else if (ps->rem != 0) { \
             panPrint_error( #FNNAME ": unaligned write\n");  \
        } else if (ps->cursws->next) { \
             ps->cursws = ps->cursws->next; \
             ps->rem = ps->cursws->length - BYTES; \
             ps->pos = ps->cursws->data; \
             *(((TYPE*)(ps->pos))++) = CONVERTFN(x); \
        } else { \
             panPrint_error( #FNNAME ": write to empty stream\n"); \
        }  \
     }

#endif

__panStream_make_read(panStream_read_uint16, 2, u_int16_t, )
__panStream_make_read(panStream_read_uint32, 4, u_int32_t, )
__panStream_make_read(panStream_read_int16,  2, int16_t,   )
__panStream_make_read(panStream_read_int32,  4, int32_t,   )

__panStream_make_read(panStream_read_lsbfirst_uint16, 2, u_int16_t,FROM_LSBF16)
__panStream_make_read(panStream_read_lsbfirst_uint32, 4, u_int32_t,FROM_LSBF32)
__panStream_make_read(panStream_read_lsbfirst_int16,  2, int16_t,  FROM_LSBF16)
__panStream_make_read(panStream_read_lsbfirst_int32,  4, int32_t,  FROM_LSBF32)

__panStream_make_read(panStream_read_msbfirst_uint16, 2, u_int16_t,FROM_MSBF16)
__panStream_make_read(panStream_read_msbfirst_uint32, 4, u_int32_t,FROM_MSBF32)
__panStream_make_read(panStream_read_msbfirst_int16,  2, int16_t,  FROM_MSBF16)
__panStream_make_read(panStream_read_msbfirst_int32,  4, int32_t,  FROM_MSBF32)

#define panStream_read_net_uint16 panStream_read_msbfirst_uint16
#define panStream_read_net_uint32 panStream_read_msbfirst_uint32
#define panStream_read_net_int16  panStream_read_msbfirst_int16
#define panStream_read_net_int32  panStream_read_msbfirst_int32

__panStream_make_write(panStream_write_uint16, 2, u_int16_t, )
__panStream_make_write(panStream_write_uint32, 4, u_int32_t, )
__panStream_make_write(panStream_write_int16,  2, int16_t,   )
__panStream_make_write(panStream_write_int32,  4, int32_t,   )

__panStream_make_write(panStream_write_lsbfirst_uint16, 2, u_int16_t,TO_LSBF16)
__panStream_make_write(panStream_write_lsbfirst_uint32, 4, u_int32_t,TO_LSBF32)
__panStream_make_write(panStream_write_lsbfirst_int16,  2, int16_t,  TO_LSBF16)
__panStream_make_write(panStream_write_lsbfirst_int32,  4, int32_t,  TO_LSBF32)

__panStream_make_write(panStream_write_msbfirst_uint16, 2, u_int16_t,TO_MSBF16)
__panStream_make_write(panStream_write_msbfirst_uint32, 4, u_int32_t,TO_MSBF32)
__panStream_make_write(panStream_write_msbfirst_int16,  2, int16_t,  TO_MSBF16)
__panStream_make_write(panStream_write_msbfirst_int32,  4, int32_t,  TO_MSBF32)

#define panStream_write_net_uint16 panStream_write_msbfirst_uint16
#define panStream_write_net_uint32 panStream_write_msbfirst_uint32
#define panStream_write_net_int16  panStream_write_msbfirst_int16
#define panStream_write_net_int32  panStream_write_msbfirst_int32
