/* $Header: /home/dl2/nygren/cvs/pan/src/common/sysconfig.linux.h,v 1.8 1998/01/18 00:11:52 nygren Exp $
 *
 * FILE:    sysconfig.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * All values are currently hard-coded.  This will be replaced with
 * something autoconf generated at some point.
 * Currently configured for gcc and OpenBSD.
 *
 */

#ifndef COMMON_SYSCONFIG_H_INC
#define COMMON_SYSCONFIG_H_INC

#include "pan-copyright.h"

#ifdef __KERNEL__

#ifndef __INCLUDE_KERNEL_VERSION__
#define __NO_VERSION__
extern char kernel_version[];
#endif
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/errno.h>

#include <linux/timer.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/bios32.h>
#include <asm/processor.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/atomic.h>

#include <sys/types.h>
/* Must define u_int{8,16,32}_t and int{8,16,32}_t */
#include <endian.h>
#include <asm/byteorder.h>

/*******************************************************/
#else /****************** !__KERNEL__ *******************/
/*******************************************************/

#include <sys/types.h>

/* Must define u_int{8,16,32}_t and int{8,16,32}_t */
#include <endian.h>
#include <asm/byteorder.h>

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <setjmp.h>

#endif /* __KERNEL__ */


#ifdef __KERNEL__

/* PAN_DEBUG is specified in GNUmakefile.global */
#define panPrint_debug(fmt,arg...) if (PAN_DEBUG) \
                                          printk(KERN_DEBUG "pan: " fmt,##arg)
#define panPrint_error(fmt,arg...) printk(KERN_ERR     "pan: " fmt,##arg)
#define panPrint_warn(fmt,arg...)  printk(KERN_WARNING "pan: " fmt,##arg)
#define panPrint_msg(fmt,arg...)   printk(KERN_INFO    "pan: " fmt,##arg)

extern void *panMemory_alloc(size_t size);
extern void *panMemory_calloc(size_t nmemb, size_t size);
#define panMemory_allocStruct(type) ((type*)panMemory_alloc(sizeof(type)))
#define panMemory_realloc realloc
extern void panMemory_free(void *ptr);

#include <jmp_buf.h>
typedef __jmp_buf jmp_buf;
extern int  panJump_setjmp(jmp_buf env);
extern void panJump_longjmp(jmp_buf env, int val);

#define assert(predicate) \
              if (!(predicate)) \
                    panPrint_error("assertion failure on line %d of %s\n", \
				   __LINE__, __FILE__)

/* Both of these must be used inside of new blocks as
 * the first defines a local variable */
#define panCritRegion_ENTER()   unsigned long procflags; \
				  save_flags(procflags); cli()
#define panCritRegion_LEAVE()   restore_flags(procflags)
#define panAtomic_dec(x) atomic_dec(x);					    



#else /* !defined(__KERNEL__) */

/* PAN_DEBUG is specified in GNUmakefile.global */
#define panPrint_debug if (PAN_DEBUG) printf
#define panPrint_error printf
#define panPrint_warn  printf
#define panPrint_msg   printf

#define panMemory_calloc calloc
#define panMemory_alloc malloc
#define panMemory_allocStruct(type) ((type*)panMemory_alloc(sizeof(type)))
#define panMemory_realloc realloc
#define panMemory_free  free
#define panJump_setjmp(e) setjmp((e))
#define panJump_longjmp(e,v) longjmp((e),(v))

/* Both of these must be used inside of new blocks as
 * the first defines a local variable */
#define panCritRegion_ENTER()  (void)0
#define panCritRegion_LEAVE()  (void)0


#endif /* !__KERNEL__ */


#define ADDR_ALIGN_MASK (0x03)
/* Returns 0 iff an addr ad 4-byte word aligned */
#define ADDR_NOT_ALIGNED_P(x) (((unsigned long)(x)) & ADDR_ALIGN_MASK)

/*
 * Definitions for byte order, according to byte significance from low
 * address to high.
 *
 *  LITTLE_ENDIAN   1234     LSB first: i386, vax 
 *  BIG_ENDIAN      4321     MSB first: 68000, ibm, net 
 *  PDP_ENDIAN      3412     LSB first in word, MSW first in int32_t 
 */

#if BYTE_ORDER == LITTLE_ENDIAN  /* LSB first: i386, vax */
#define FROM_LSBF16(x) (x)
#define FROM_LSBF32(x) (x)
#define TO_LSBF16(x)   (x)
#define TO_LSBF32(x)   (x)
#define FROM_MSBF16(x) ntohs(x)
#define FROM_MSBF32(x) ntohl(x)
#define TO_MSBF16(x)   ntohs(x)
#define TO_MSBF32(x)   ntohl(x)
#else
#error "unsupported byte order "
#endif

/* Do sockets have a sin_len field? */
#undef HAVE_SIN_LEN

/* Number of microseconds to which gettimeofday is accurate */
#define panTimestamp_USEC_RESOLUTION 5  /* Actually less, but the system
					 * call takes time... */

#endif /* COMMON_SYSCONFIG_H_INC */
