/* $Header: /home/dl2/nygren/cvs/pan/src/common/sysconfig.openbsd.h,v 1.5 1998/01/18 00:11:53 nygren Exp $
 *
 * FILE:    sysconfig.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * All values are currently hard-coded.  This will be replaced with
 * something autoconf generated at some point.
 * Currently configured for gcc and OpenBSD.
 *
 */

#ifndef COMMON_SYSCONFIG_H_INC
#define COMMON_SYSCONFIG_H_INC

#include "pan-copyright.h"

#include <sys/types.h>
#include <machine/endian.h>

/* Must define u_int{8,16,32}_t and int{8,16,32}_t */
#include <machine/types.h>

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* PAN_DEBUG is specified in GNUmakefile.global */
#define panPrint_debug if (PAN_DEBUG) printf
#define panPrint_error printf
#define panPrint_warn  printf
#define panPrint_msg   printf

#define panMemory_calloc calloc
#define panMemory_alloc malloc
#define panMemory_allocStruct(type) ((type*)panMemory_alloc(sizeof(type)))
#define panMemory_realloc realloc
#define panMemory_free  free

#define ADDR_ALIGN_MASK (0x03)
/* Returns 0 iff an addr ad 4-byte word aligned */
#define ADDR_NOT_ALIGNED_P(x) (((unsigned long)(x)) & ADDR_ALIGN_MASK)

/*
 * Definitions for byte order, according to byte significance from low
 * address to high.
 *
 *  LITTLE_ENDIAN   1234     LSB first: i386, vax 
 *  BIG_ENDIAN      4321     MSB first: 68000, ibm, net 
 *  PDP_ENDIAN      3412     LSB first in word, MSW first in int32_t 
 */

#if BYTE_ORDER == LITTLE_ENDIAN  /* LSB first: i386, vax */
#define FROM_LSBF16(x) (x)
#define FROM_LSBF32(x) (x)
#define TO_LSBF16(x)   (x)
#define TO_LSBF32(x)   (x)
#define FROM_MSBF16(x) __byte_swap_int16(x)
#define FROM_MSBF32(x) __byte_swap_int32(x)
#define TO_MSBF16(x)   __byte_swap_int16(x)
#define TO_MSBF32(x)   __byte_swap_int32(x)
#elif BYTE_ORDER == BIG_ENDIAN  /* MSB first: 68000, ibm, net */
#define FROM_LSBF16(x) __byte_swap_int16(x)
#define FROM_LSBF32(x) __byte_swap_int32(x)
#define TO_LSBF16(x)   __byte_swap_int16(x)
#define TO_LSBF32(x)   __byte_swap_int32(x)
#define FROM_MSBF16(x) (x)
#define FROM_MSBF32(x) (x)
#define TO_MSBF16(x)   (x)
#define TO_MSBF32(x)   (x)
#else
#error "unsupported byte order "
#endif

/* Do sockets have a sin_len field? */
#define HAVE_SIN_LEN

/* Number of microseconds to which gettimeofday is accurate */
#define panTimestamp_USEC_RESOLUTION 10000

#endif /* COMMON_SYSCONFIG_H_INC */
