/* $Header: /home/dl2/nygren/cvs/pan/src/common/timestamp.c,v 1.2 1998/01/20 23:44:18 nygren Exp $
 *
 * FILE:    timestamp.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Routines for getting timing information from a node.  Node that times
 * are particular to individual nodes and may not be synchronized
 * between nodes.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swsstream.h>
#include <common/timestamp.h>

/* Reads a time stamp from a stream */
int
panTimestamp_read(panTimestamp *ts, panStream *ps)
{  
  ts->sec  = panStream_read_net_int32(ps);
  ts->usec = panStream_read_net_int32(ps);
  return 8; /* TODO: Should test for success... */
}

/* Writes a time stamp from a stream */
int
panTimestamp_write(panStream *ps, panTimestamp *ts)
{
  panStream_write_net_int32(ps, ts->sec);
  panStream_write_net_int32(ps, ts->usec);
  return 8; /* TODO: Should test for success... */
}

/* Generates a string from a time stamp and puts it into buf.
 * The format is determined by the oring together some of the above.
 * If buf is NULL, returns safebuf (which is NOT threadsafe and 
 * will change the next time stringify is called).
 */
char *
panTimestamp_stringify(char *buf, int buflen, panTimestamp *ts, int fmt)
{
  panTimestamp t;
  static char safebuf[panTimestamp_STRINGIFY_BUFSIZE];
  char *separator = (fmt&panTimestamp_STRINGIFY_FMT_INSECS)?".":"";
  t.sec = ts->sec; 
  t.usec = ts->usec; 
  __panTimestamp_normalize(&t);
  if ((fmt&panTimestamp_STRINGIFY_FMT_PADSECS)
      && (fmt&panTimestamp_STRINGIFY_FMT_PADUSECS))
    {
      sprintf(safebuf,"%10d%s%06d", t.sec, separator, t.usec);
    }
  else if (!(fmt&panTimestamp_STRINGIFY_FMT_PADSECS)
	   && (fmt&panTimestamp_STRINGIFY_FMT_PADUSECS))
    {
      sprintf(safebuf,"%d%s%06d", t.sec, separator, t.usec);
    }
  else if ((fmt&panTimestamp_STRINGIFY_FMT_PADSECS)
	   && !(fmt&panTimestamp_STRINGIFY_FMT_PADUSECS))
    {
      if (fmt&panTimestamp_STRINGIFY_FMT_INSECS)
	{
	  if (t.usec)
	    sprintf(safebuf,"%10d.%06d", t.sec, t.usec);
	  else
	    sprintf(safebuf,"%10d", t.sec);
	}
      else
	{
	    sprintf(safebuf,"%10d%06d", t.sec, t.usec);
	}
    }
  else /* if (!(fmt&panTimestamp_STRINGIFY_FMT_PADSECS)
	   && !(fmt&panTimestamp_STRINGIFY_FMT_PADUSECS)) */
    {
      if (fmt&panTimestamp_STRINGIFY_FMT_INSECS)
	{
	  if (t.usec)
	    sprintf(safebuf,"%d.%06d", t.sec, t.usec);
	  else
	    sprintf(safebuf,"%d", t.sec);
	}
      else
	{
	  if (t.sec)
	    sprintf(safebuf,"%d%06d", t.sec, t.usec);
	  else
	    sprintf(safebuf,"%d", t.usec);
	}
    }
  if (buf)
    {
      strncpy(buf, safebuf, buflen-1);
      buf[buflen-1]='\0';
      return buf;
    }
  else
    {
      return safebuf;
    }
}

/* Returns 0 if equal, -1 if ts1<ts2, 1 if ts1>ts2.
 * Also has the effect of normalizing ts1 and ts2. */
int
panTimestamp_compare(panTimestamp *ts1, panTimestamp *ts2)
{
  __panTimestamp_normalize(ts1);
  __panTimestamp_normalize(ts2);
  if (ts1->sec>ts2->sec)
    return 1;
  else if (ts1->sec<ts2->sec)
    return -1;
  else if (ts1->usec>ts2->usec)
    return 1;
  else if (ts1->usec<ts2->usec)
    return -1;
  else 
    return 0;
}

/* Subtracts one timestamp from another and puts the result in result. */
void
panTimestamp_subtract(panTimestamp *result, 
		      panTimestamp *ts1, panTimestamp *ts2)
{
  result->sec = ts1->sec - ts2->sec;
  result->usec = ts1->usec - ts2->usec;
  __panTimestamp_normalize(result);
}

/* Make usec be a value in the range of 0 to 999999 */
void
panTimestamp_normalize(panTimestamp *ts)
{
  __panTimestamp_normalize(ts);
}


