/* $Header: /home/dl2/nygren/cvs/pan/src/common/timestamp.h,v 1.1 1998/01/18 00:11:54 nygren Exp $
 *
 * FILE:    timestamp.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Routines for getting timing information from a node.  Node that times
 * are particular to individual nodes and may not be synchronized
 * between nodes.
 *
 */

#ifndef NODE_TIMESTAMP_H_INC
#define NODE_TIMESTAMP_H_INC

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swsstream.h>

/* Should be identical to a timeval! 
 * Maybe it should really be a timespec?  (with nanoseconds instead?) */
typedef struct _panTimestamp {
  int32_t  sec;			/* Number of seconds */
  int32_t  usec;		/* Number of microseconds */
} panTimestamp;



extern int
panTimestamp_read(panTimestamp *ts, panStream *ps);

extern int
panTimestamp_write(panStream *ps, panTimestamp *ts);

#define panTimestamp_STRINGIFY_FMT_INSECS   0x01
#define panTimestamp_STRINGIFY_FMT_INUSECS  0x00
#define panTimestamp_STRINGIFY_FMT_PADUSECS 0x02
#define panTimestamp_STRINGIFY_FMT_PADSECS  0x04
#define panTimestamp_STRINGIFY_BUFSIZE  (11+6+2)
extern char *
panTimestamp_stringify(char *buf, int buflen, panTimestamp *ts, int format);

extern int
panTimestamp_compare(panTimestamp *ts1, panTimestamp *ts2);

extern void
panTimestamp_subtract(panTimestamp *result, 
		      panTimestamp *ts1, panTimestamp *ts2);

extern void
panTimestamp_normalize(panTimestamp *ts);

/* These functions are implemented in machine-specific code in the
 * arch directory... */

extern int
panTimestamp_getCurrent(panTimestamp *ts);

extern int
panTimestamp_getResolution(panTimestamp *res);


#define panTimestamp_dup(dst, src) \
        memcpy((void*)(dst), (void*)(src), sizeof(panTimestamp))

/* Make usec be a value in the range of 0 to 999999 */
extern inline void
__panTimestamp_normalize(panTimestamp *ts)
{
  if (ts->usec>=1000000)
    {
      ts->sec += ((ts->usec - (ts->usec%1000000)))/1000000;
      ts->usec = (ts->usec%1000000);
    }
  if (ts->usec<0)
    {
      ts->sec  -= ((-ts->usec)-((-ts->usec)%1000000))/1000000;
      ts->usec  = -((-ts->usec)%1000000); 
      if (ts->usec<0)
	{
	  ts->usec  = 1000000+ts->usec;
	  ts->sec--;
	}
    }
}

#endif /* NODE_TIMESTAMP_H_INC */
