# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgAccessSpec.pm,v 1.1 1997/11/14 07:24:02 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgAccessSpec.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgObject derivative for holding an access specifier.
# An access specifier specifies how a certain block of generated code
# will deal with access to a field in a structure.  For example,
# a read-only fields will be skipped and not filled in during a read.
# An access spec inclides a specifier (corresponding to the "spec" field
# specified by the active generate block) and the access mode which
# may be one of:
#
#      ro = read-only
#      wo = write-only
#      rw = read-write (the default)
#      na = no access (always skip -- no reads or writes) 
#

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;

package psgAccessSpec;
@ISA = qw( psgObject );


# Passed a specifier (an identifier Terminal) 
# and an access mode (another terminal).
sub new {
    my ($type, $spec, $amode) = @_;
    my $self = psgObject->new($spec->getLocation());
    $self->{spec} = $spec;
    $self->{amode} = $amode;    
    return bless $self, $type;
}

# Returns a string containing the access specifier
sub getSpecifier {
    my ($self) = @_;
    return $self->{spec}->getTermVal();
}

# Returns a string containing the access mode
sub getAccessMode {
    my ($self) = @_;
    return $self->{amode}->getTermVal();
}

sub isReadable {
    my ($self) = @_;
    if (($self->getAccessMode eq "ro")
	|| ($self->getAccessMode eq "rw")) {
	return 1;
    } else {
	return 0;
    }
}

sub isWritable {
    my ($self) = @_;
    if (($self->getAccessMode eq "wo")
	|| ($self->getAccessMode eq "rw")) {
	return 1;
    } else {
	return 0;
    }
}

sub getDebug {
    my ($self) = @_;
    return "(aspec:".$self->getLineNumber." ".$self->getSpecifier." ".$self->getAccessMode.")";} 

