# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgConstDef.pm,v 1.1 1997/11/14 07:24:03 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgConstDef.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgObject derivative that contains a const (constant) definition
#

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;

package psgConstDef;
@ISA = qw( psgObject );

# Creates a new constant definition from an identifier terminal and 
# a psgNumVal.
sub new {
    my ($type, $ident, $numval) = @_;

    return $ident  if ($ident->isError);
    return $numval if ($numval->isError);
    my $self = psgObject->new($ident->getLocation);
    $self->{c_name} = $ident;
    $self->{c_val} = $numval;
    return bless $self, $type;
}

# returns an identifier terminal
sub getNameIdent {
    my ($self) = @_;
    return $self->{c_name};
}

# returns a string
sub getName {
    my ($self) = @_;
    return $self->{c_name}->getTermVal;
}

sub getPrefixedName {
    my ($self, $gen) = @_;
    my $pname = $gen->getPrefix.$self->getName;
    return $pname;
}

# returns a NumVal
sub getConstVal {
    my ($self) = @_;
    return $self->{c_val};
}

# Writes the constant definition out to a generator
sub makeConstDef {
    my ($self, $gen) = @_;
    $gen->putLine("#define ".$self->getPrefixedName($gen)
		  ."\t\t".$self->getConstVal->getValueString);
}



sub getDebug {
    my ($self) = @_;
    return "(constdef:".$self->getLineNumber." ".$self->getNameIdent->getDebug." ".$self->getConstVal->getDebug.")";
} 

