# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgFieldList.pm,v 1.1 1997/11/14 07:24:04 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgFieldList.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgObjList derivative that contains a list of psgFields.
#

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;
use psgObjList;
use psgField;
use psgType;

package psgFieldList;
@ISA = qw( psgObjList );

# location contains the location number of the terminal or non-terminal
sub new {
    my ($type, $location) = @_;
    my $self = psgObjList->new($location);
    return bless $self, $type;
}

# Most methods are inherited from psgObjList

# Adds a type to all elements in the list and returns self
sub setType {
    my ($self, $newtype, $symtab) = @_;
    for $i (0..(@{$self->{listelems}}-1)) {
	$self->{listelems}->[$i] 
	    = $self->{listelems}->[$i]->setType($newtype, $symtab);
    }
    return $self;
}
 
# Adds a list of access specs to all elements in the list and returns self
sub setAccessSpecs {
    my ($self, $newaspecs) = @_;
    for $i (0..(@{$self->{listelems}}-1)) {
	$self->{listelems}->[$i] 
	    = $self->{listelems}->[$i]->setAccessSpecs($newaspecs);
    }
    return $self;
}

# Returns the total size of a fieldlist and tells each field
# its offset from the start of the structure.
sub getTotalSizeAndSetFieldPos {
    my ($self) = @_;
    my @elems = $self->getElements;
    my $curpos = 0;
    foreach $el (@elems) {
	$curpos = $el->getType->getStartPos($curpos);
	$el->setPos($curpos);
	$curpos += $el->getSize;
    }
    return $curpos;
}


sub makeStructContents {
    my ($self, $gen) = @_;    
    my @elems = $self->getElements;
    my $curpos = 0;
    foreach $el (@elems) {
	$el->makeStructContents($gen);
    }
}


 
sub getDebug {
    my ($self) = @_;
    $s = "(fieldlist:".$self->getLineNumber." "; $first=0;
    for $i (@{$self->{listelems}}) {
	if ($first==0) { $first=1; }
	else { $s .= ","; }
	$s .= $i->getDebug();
    }
    $s .= ")";
}
