#!/usr/bin/perl
#
# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgMain.pl,v 1.2 1997/12/01 22:16:07 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgMain.pl
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# Mainline for PAN Serializer Generator
#


push(@INC,  ("../parselex"));

use Parse::CLex;
use psgParser;
use psgGenerator;
use psgError;
use Getopt::Long;

package psgMain;

$opt_verbose           = 0;   # verbosity level


sub print_usage {
    my ($errmsg) = @_;
    print "\n";
    print "Usage:\n";
    print "        panSerGen [-I dir]+ [--input-search-path=dir]\n";
    print "                  [-v] [--debug=verbosity-level]\n";
    print "                  [--help] [-h]\n";
    print "                  [--output-directory=dir] [-O dir]\n";
    print "                  input-file.psg [input-file.psg]+\n";
    print "\n";
    if ($errmsg eq "help") {
	exit 0;
    }
    if ($errmsg) {
	die $errmsg;
    }
}

# Searches the input path and returns a file handle reference and the
# full pathname if the file could be found
sub search_and_open {
    my ($fname, @spath) = @_;    

    if ($fname eq "-") {
	return \*STDIN, "<stdin>";
    }
    foreach $pathelem (@spath) {
	print "Trying to open $pathelem/$fname\n" if ($opt_verbose>5);
	if (open INFILE, "<$pathelem/$fname") {
	    return \*INFILE, $pathelem."/".$fname;
	}
    }
    print "Trying to open $fname\n" if ($opt_verbose>5);
    if (open INFILE, "<$fname") {
	return \*INFILE, $fname;
    }
    return 0;
}

# Parses @ARGV, loads in the specified input files,
# then generates the specified output files.
sub psgMain {

    my ($lexer, $p);
    local @opt_input_search_path = ();
    local $opt_output_dir        = "";  # prefix to output filenames
    local @opt_input_files;
    my ($opt_dash, $opt_v);

    if (0 == &Getopt::Long::GetOptions
	                ("debug:i"                => \$opt_verbose,
			 "v"                      => \$opt_v,
			 ""                       => \$opt_dash,
			 "input-search-path|I:s@" => \@opt_input_search_path,
			 "help|h"                 => sub{print_usage("help");},
			 "output-directory|O:s"   => \$opt_output_dir)) {
	print_usage("panSerGen: *** Passed invalid argument\n");
    }
    if ($opt_v && !$opt_verbose) { $opt_verbose = 1; }
    print "Verbosity set to $opt_verbose\n" if ($opt_verbose>4);
    if ($opt_dash) { push @ARGV, "-"; }

    if (!@ARGV) {
	print_usage("panSerGen: *** No input files specified\n");
    }

    @opt_input_files = @ARGV;

    foreach $infilename (@opt_input_files) {
	my ($infile, $realinname);
	
	($infile, $realinname) = search_and_open($infilename, 
						 @opt_input_search_path);
	if (!$infile) { die "panSerGen: *** Couldn't find $infilename\n"; }
	print "Loading $realinname...\n" if ($opt_verbose>0);

	$lexer = psgParser::newLexer($infile);
	$lexer->fromName($realinname);
	$p = psgParser->new(\&psgParser::yylex, \&psgParser::yyerror, 0);
	
	my ($generators, $constsandstructs, $symtab) = $p->doParse($lexer);
	
	print "Done parsing $realinname...\n" if ($opt_verbose>1);

	if ($constsandstructs->containsError || $generators->containsError) {
	    die "panSerGen: *** Input file contains errors.  No output generated.\n";
	}
	foreach $gen ($generators->getElements) {
	    print "  Making ".$gen->getOutFilename."...\n" if ($opt_verbose>0);
	    $gen->makeOutputFile($constsandstructs);	    
	}
	print "Done with $realinname...\n" if ($opt_verbose>1);
    }
}


psgMain();
