# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgNumVal.pm,v 1.1 1997/11/14 07:24:05 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgNumVal.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgObject derivative that contains a numeric value.
# May come either from a number or from a constant.
#

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;
use psgError;

package psgNumVal;
@ISA = qw( psgObject );

# Passed a value source and a numeric value.
sub new {
    my ($type, $valsource, $numval) = @_;
    my $self = psgObject->new($valsource->getLocation);
    return $valsource if ($valsource->isError);
    $self->{nv_source} = $valsource;
    $self->{nv_value} = $numval;
    return bless $self, $type;
}

# Passed an identifier terminal and a symbol table 
sub newFromIdentifier {
    my ($type, $valsource, $symtab) = @_;
    return $valsource if ($valsource->isError);

    my $symres = $symtab->getSym($valsource);
    if ($symres && (ref($symres) eq psgConstDef)) {
	return $type->new($valsource, $symres->getConstVal->getValue);
    } else {
	my $errmsg = "couldn't resolve ".$valsource->getTermVal;
	$valsource->printError($errmsg);
	return psgError->new($valsource, $errmsg);
    }
}

# Passed a number terminal
sub newFromNumber {
    my ($type, $valsource) = @_;
    return $type->new($valsource, $valsource->getTermVal);
}

sub getValueString {
    my ($self) = @_;
    return $self->{nv_value};
}

sub getValue {
    my ($self) = @_;
    return $self->{nv_value};
}

sub getValueSource {
    my ($self) = @_;
    return $self->{nv_source};
}

sub getDebug {
    my ($self) = @_;
    return "(numval:".$self->getLineNumber." ".$self->getValue.")";
} 

