# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgObjList.pm,v 1.1 1997/11/14 07:24:06 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgObjList.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgObject derivative that contains a list of psgObjects.
# Useful to allow items in a list to be rapidly accumulated during parsing.
#

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;

package psgObjList;
@ISA = qw( psgObject );

# location contains the location number of the terminal or non-terminal

sub new {
    my ($type, $location) = @_;
    my $self = psgObject->new($location);
    $self->{listelems} = [];
    return bless $self, $type;
}

sub newWithElem {
    my ($type, $firstelem) = @_;
    my $newlist = $type->new($firstelem->getLocation());
    $newlist->addElement($firstelem);
    return $newlist;
}

sub newWithList {
    my ($type, $initelemlist) = @_;
    my $newlist = $type->new($initelemlist->getLocation());
    $newlist->appendList($initelemlist);
    return $newlist;
}

sub containsError {
    my ($self) = @_;
    foreach $el (@{$self->{listelems}}) {
	if ($el->isError) { return 1; }
    }
    return 0;
}

sub getElements {
    my ($self) = @_;
    return @{$self->{listelems}};
}

# Adds an element to the front of the list and returns self
sub addElement {
    my ($self, $newelem) = @_;
    unshift @{$self->{listelems}}, $newelem;
#    print "Added ".($newelem->getDebug())." to objlist with ".(@{$self->{listelems}})." elements now ".($self->getDebug())."\n";
    
    return $self;
}
 
# Appends an ObjList to the list and returns self
sub appendList {
    my ($self, $newlist) = @_;
    unshift @{$self->{listelems}}, $newlist->getElements;
#    print "Added ".($newlist->getDebug())." to objlist with ".(@{$self->{listelems}})." elements now ".($self->getDebug())."\n";
    
    return $self;
}
 
sub getDebug {
    my ($self) = @_;
    $s = "(objlist:".$self->getLineNumber." "; $first=0;
    for $i (@{$self->{listelems}}) {
	if ($first==0) { $first=1; }
	else { $s .= ","; }
	$s .= $i->getDebug();
    }
    $s .= ")";
}
