# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgObject.pm,v 1.1 1997/11/14 07:24:06 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgObject.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# Base class object for the PAN Serializer Generator.
# Contains methods for retrieving the line number and
# other things which may be overridden by derived classes.
#

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgLocation;

package psgObject;

# lineno contains the line number of the terminal or non-terminal

sub new {
    my ($type, $loc) = @_;
    my $self = {};
    $self->{loc} = $loc;
    return bless $self, $type;
}

sub getLocation {
    my ($self) = @_;
    return $self->{loc};
}

sub getLineNumber {
    my ($self) = @_;
    return $self->{loc}->getLineNumber;
}

sub getFilename {
    my ($self) = @_;
    return $self->{loc}->getFilename;
}

sub getDebug {
    my ($self) = @_;
    return "(object:".$self->getLineNumber.")";
}

sub isError {
    my ($self) = @_;
    return 0;
}

sub printWarn {
    my ($self, $msg) = @_;
    return $self->{loc}->printWarn($msg);
}

sub printError {
    my ($self, $msg) = @_;
    return $self->{loc}->printError($msg);
}

