%{

# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgParser.y,v 1.2 1997/12/01 22:16:08 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgParser.y
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# Contains a yacc description of a parser for the PAN Serializer Generator.
# Compile this file using the version of Berkeley byacc that is capable
# of generating OO perl5 code (using the -P option) then rename the
# file to psgParser.pm.
#
###########################################################################


require 5;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgTerminal; 
use psgObjList; 
use psgTaggedVal; 
use psgObject; 
use psgField; 
use psgFieldList; 
use psgConstDef; 
use psgLocation; 
use psgSymTab; 
use psgType; 
use psgStructDef; 
use psgNumVal; 
use psgTVList; 
use psgGenerator; 
use psgAccessSpec; 

package psgParser;
%}

/* Tokens with values */
%token STRING IDENTIFIER LITEXPR NUMBER

/* Keywords */
%token STRUCT GENERATE CONST
%token SPEC GENTYPE MAKE PREFIX
%token RO WO RW NA
%token READCODE WRITECODE READPROTO WRITEPROTO
%token ';' '{' '}' ':' ',' '[' ']'

%%

expr:   
    generate_list top_level_decl_list     { $$ = [ $1, $2 ]; }
    ;

generate_list: 
    generate_statement                    { $$ = psgObjList->newWithElem($1); }
    | generate_statement generate_list    { $$ = $2->addElement($1); }
    ;

top_level_decl_list:
    top_level_decl                  	  { $$ = psgTVList->newWithElem($1); }
    | top_level_decl top_level_decl_list  { $$ = $2->addElement($1); }
    ;

top_level_decl:
    const_def     { $$ = psgTaggedVal->new("const",  $1, $1); }
    | struct_def  { $$ = psgTaggedVal->new("struct", $1, $1); }
    ;

const_def: 
    CONST IDENTIFIER numeric_value ';'  { $$ = psgConstDef->new($2, $3);
	       			  	  $ctxsym->addSym($2, $$); }
    ;

numeric_value:
    NUMBER         { $$ = psgNumVal->newFromNumber($1); } 
    | IDENTIFIER   { $$ = psgNumVal->newFromIdentifier($1, $ctxsym); }
    ;

struct_def:
    STRUCT IDENTIFIER '{' field_list '}' ';' 
	{ $$ = psgStructDef->new($2, $4); 
	  $ctxsym->addSym($2, $$); }
    ;

field_list:
    component_decl                { $$ = psgFieldList->newWithList($1); }
    | component_decl field_list   { $$ = $2->appendList($1); }
    ;

component_decl:  /* Returns the modified declarator list */
    access_specifier_list type_specifier declarator_list ';'
        { $$ = $3->setAccessSpecs($1)->setType($2, $ctxsym); }	    
    | type_specifier declarator_list ';'  
        { $$ = $2->setType($1, $ctxsym); }	    
    ;

declarator_list:
    declarator				{ $$ = psgFieldList->newWithElem($1); }
    | declarator ',' declarator_list	{ $$ = $3->addElement($1); }
    ;

declarator:   /* Actually returns a psgField or a psgError */
    IDENTIFIER				          { $$ = psgField->new($1); }
    | IDENTIFIER '[' numeric_value ']'  
				{ $$ = psgField->new($1)->setArraySize($3); }
    ;

access_specifier_list:
    access_specifier                    { $$ = psgObjList->newWithElem($1); }
    | access_specifier ',' access_specifier_list { $$ = $3->addElement($1); }
    ;

type_specifier:	      /* May eventually include structures */
    IDENTIFIER        { $$ = $1; }   
    ;

access_mode:
    RO      { $$ = $1; }
    | WO    { $$ = $1; }
    | RW    { $$ = $1; }
    | NA    { $$ = $1; }
    ;

access_specifier:
    IDENTIFIER ':' access_mode	{ $$ = psgAccessSpec->new($1, $3); }
    ;

generate_statement:
    GENERATE STRING '{' gen_mod_list '}' ';'
    		{ $$ = psgGenerator->newWithModList($2, $4); }
    | GENERATE STRING gen_type ';' 
    		{ $$ = psgGenerator->newWithDefaults($2, $3); }
    ;

gen_type:     /* one of capsule, appheader, appcode */
    IDENTIFIER                { $$ = $1; }
    ;

gen_mod_list:
    gen_mod                   { $$ = psgTVList->newWithElem($1); }
    | gen_mod gen_mod_list    { $$ = $2->addElement($1); }
    ;

gen_mod:
    SPEC IDENTIFIER ';'       { $$ = psgTaggedVal->new("spec",    $2, $1); }
    | GENTYPE gen_type ';'    { $$ = psgTaggedVal->new("gentype", $2, $1); }
    | MAKE gen_make_list ';'  { $$ = psgTaggedVal->new("make",    $2, $1); }
    | PREFIX STRING ';'       { $$ = psgTaggedVal->new("prefix",  $2, $1); }
    ;

gen_make_list:
    gen_make                       { $$ = psgObjList->newWithElem($1); }
    | gen_make ',' gen_make_list   { $$ = $3->addElement($1); }
    ;

gen_make:
    STRUCT         { $$ = $1; }  /* All just return a psgTerminal */
    | CONST        { $$ = $1; }
    | READCODE     { $$ = $1; }
    | WRITECODE    { $$ = $1; }
    | READPROTO    { $$ = $1; }
    | WRITEPROTO   { $$ = $1; }
    ;


/*
oldexpr:		NUMBER		{ $$ = $1; }
	|	expr '+' expr	{ $$ = $1 + $3; }
	|	expr '-' expr	{ $$ = $1 - $3; }
	|	expr '*' expr	{ $$ = $1 * $3; }
	|	expr '/' expr   { $$ = $1 / $3; }
        |       TWO             { $$ = 2; }
	|	'(' expr ')'	{ $$ = $2; }
	;
*/
%%

# Actually does the parsing.  Creates a dynamic parsing context
# for the symbol table.  Returns both what yyparse returned and
# the symbol table.
sub doParse {
    my ($p, $s) = @_;

    local $ctxsym = psgSymTab->new();  # dynamically scoped

    $ctxsym->addInitial;             # add initial entries
    $ctxsym->printDebug if ($psgMain::opt_verbose>6);    
    $pret = $p->yyparse($s);
    $ctxsym->printDebug if ($psgMain::opt_verbose>6);  ;
    if ($psgMain::opt_verbose>6) { print "PRET   = ".$pret."\n"; }
    if ($psgMain::opt_verbose>6) { print "CTXSYM = ".$ctxsym."\n"; }
    return (@{$pret}, $ctxsym);
}


@lexertokens = (
	   qw(tCOMMENT),     [qw(/\* (?:.|\n)*(?=\*/) \*/)],

	   qw(
	      tSINGLECHAR    [;{}:,\[\]]
	      tSTRUCT        \bstruct\b
	      tGENERATE      \bgenerate\b
	      tCONST         \bconst\b
	      tGENTYPE       \bgentype\b
	      tSPEC          \bspec\b
	      tMAKE          \bmake\b
	      tPREFIX        \bprefix\b
	      tRO            \bro\b
	      tWO            \bwo\b
	      tRW            \brw\b
	      tNA            \bna\b
	      tREADCODE      \breadcode\b
	      tWRITECODE     \bwritecode\b
	      tREADPROTO     \breadproto\b
	      tWRITEPROTO    \bwriteproto\b
	      tNUMBER        [1-9][0-9]*
	      tNEWLINE       \n
	      tIDENTIFIER    [a-zA-Z_][a-zA-Z_0-9]*
	      ),

	   qw(tSTRING),   [qw(" (?:[^"]+|"")* ")],
#"
	   qw(tERROR  .*), sub {
	       die qq!can\'t analyze: "$_[1]"!;
	   }
	 );
#print "LEXERTOKENS: <".join ("~", @lexertokens).">\n\n";

sub newLexer
{
    my ($source) = @_;
    my $lexer = Parse::CLex->new(@lexertokens);
    $lexer->skip("[ \t]+");
    Parse::CLex->trace if ($psgMain::opt_verbose>8);
    if ($psgMain::opt_verbose>6) 
	{ printf "Source: <%s>, %d\n",$source,fileno($source)};
    $lexer->from($source);
    if ($psgMain::opt_verbose>6) 
	{ printf "Skip set to <%s> \n", $lexer->skip;}

    return $lexer;
}

sub yylex
{
    my ($lexer) = @_;
    my ($tokname, $token);

    #print "About to call lexer from yylex (NUMBER=$NUMBER, tNUMBER=$tNUMBER)\n";
  GETTOKENS:
    while (1) 
    {  # Swollow comments and newlines
	$token = $lexer->next;
	if ($lexer->eoi) {
	    return 0;
	} elsif ($token->name eq tNEWLINE
		 || $token->name eq tCOMMENT) {
	    # printf "Eating %s<%s>\n", $token->name, $token->getstring;
	} else { last GETTOKENS; }
    }
    
    if ($token->name eq "tSINGLECHAR") {
	$terminal = psgTerminal->new(psgLocation->new($lexer->recordno,
						      $lexer->fromName),
				     "SINGLECHAR", 
				     $token->getstring); 
	#printf "Got ".($terminal->getDebug)."\n";
	return (ord($token->getstring), $terminal);
		   
    } elsif ($token->name eq "tSTRING") {  # Remove the quotes from the value
	$tokname = "STRING";
	$tokval = $token->getstring;
	$tokval =~ s/^"(.*)"$/$1/;
	$terminal = psgTerminal->new(psgLocation->new($lexer->recordno,
						      $lexer->fromName),
				     $tokname, $tokval);
	#printf "Got ".($terminal->getDebug)."\n";
        return (${$psgParser::{$tokname}}, $terminal);
    } else {
	$tokname = $token->name;
	$tokname =~ s/^t//;
	$terminal = psgTerminal->new(psgLocation->new($lexer->recordno,
						      $lexer->fromName),
				     $tokname, 
				     $token->getstring);
	#printf "Got ".($terminal->getDebug)."\n";
        return (${$psgParser::{$tokname}}, $terminal);
    }
}

sub yyerror {
    my ($msg, $lexer) = @_;
    die (($lexer->fromName).":".($lexer->recordno).": $msg\n");
}

1;
