# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgSymTab.pm,v 1.3 1998/01/18 00:12:01 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgSymTab.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# Contains a symbol table for the PAN Serializer Generator
#

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;
use psgType;
use psgError;
use psgWordIOBuiltIn;
use psgBuiltInType;

package psgSymTab;

sub new {
    my ($type) = @_;
    my $self = {};
    $self->{syms} = {};
    return bless $self, $type;
}

# Adds the symbol with name and value to the symbol table.
# Name should be a string or an identifier terminal.
# Val should be a psgObject or derivative.
# Doesn't add error nonterminals.
sub addSym {
    my ($self, $name, $val) = @_;
    my $namestr;

    if ($val->isError) { return; }
    if (ref $name) {
	if ($name->isError) { return; }
	$namestr = $name->getTermVal();
    } else {
	$namestr = $name; 
    }

    if ($self->{syms}->{$namestr}) {
	$val->printWarn("$namestr already defined (previous definition at ".
			$self->{syms}->{$namestr}->stringify.")");
    }

    $self->{syms}->{$namestr} = $val;

    print "addSym bound $namestr to ".$val->getDebug."\n" 
	if ($psgMain::opt_debug>5);
}

# Looks up a symbol with name in the symbol table, with
# an optional reference to a list of prefix strings.
# Name should be a string or an identifier terminal.
sub getSym {
    my ($self, $name, $prefixes) = @_;
    my $namestr;

    if (ref $name) {
	$namestr = $name->getTermVal();
    } else {
	$namestr = $name; 
    }

    if ($self->{syms}->{$p.$namestr}) {
	return $self->{syms}->{$p.$namestr};
    }

    if ($prefixes) {
	foreach $p (@$prefixes) {
	    if ($self->{syms}->{$p.$namestr}) {
		return $self->{syms}->{$p.$namestr};
	    }
	}
    }

    return 0;
}


# Adds the initial symbols
sub addInitial {
    my ($self) = @_;
    my $dummyloc = psgLocation->new(0,"<internal>");
    
    $self->addSym("CodeobjName", psgBuiltInType->new("CodeobjName", 24,", 0"));
    $self->addSym("NodeAddr",    psgBuiltInType->new("NodeAddr", 16, ""));
    $self->addSym("Timestamp",   psgBuiltInType->new("Timestamp", 8, ""));
    $self->addSym("int8",      psgWordIOBuiltIn->new(1, 8));
    $self->addSym("int16",     psgWordIOBuiltIn->new(1, 16));
    $self->addSym("int32",     psgWordIOBuiltIn->new(1, 32));
    $self->addSym("uint8",     psgWordIOBuiltIn->new(0, 8));
    $self->addSym("uint16",    psgWordIOBuiltIn->new(0, 16));
    $self->addSym("uint32",    psgWordIOBuiltIn->new(0, 32));
    $self->addSym("int8_t",    psgWordIOBuiltIn->new(1, 8));
    $self->addSym("int16_t",   psgWordIOBuiltIn->new(1, 16));
    $self->addSym("int32_t",   psgWordIOBuiltIn->new(1, 32));
    $self->addSym("u_int8_t",  psgWordIOBuiltIn->new(0, 8));
    $self->addSym("u_int16_t", psgWordIOBuiltIn->new(0, 16));
    $self->addSym("u_int32_t", psgWordIOBuiltIn->new(0, 32));
}


sub printDebug {
    my ($self) = @_;
    
    print "---- SYMTAB ----\n";
    foreach $e (keys %{$self->{syms}}) {
	print "   > $e => ".$self->{syms}->{$e}->getDebug."\n";
    }
    print "----------------\n";
}
