/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/linux2/app_if.h,v 1.2 1997/12/18 02:33:29 nygren Exp $
 *
 * FILE:    app_if.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides routines for communicating with applications connected
 * to a PAN node through AF_PAN sockets.
 *
 */

#ifndef NODE_APPIF_H_INC
#define NODE_APPIF_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <common/sysconfig.h>
#include <node/nodecodeobj.h>
#include <common/nodeaddr.h>
#include <common/conventions.h>



/* These panNodeAppLinkSocket's are very similar to unix_socket's.  In
 * fact we even use the protinfo.af_unix field of the socket
 * structure, to avoid having to make any changes to the kernel.  In
 * the unix_opt struct (which protinfo.af_unix is), we commandeer the
 * protinfo.af_unix.other pointer to instead point to a panNodeAppLink
 * (which happens to also be a swseg).  For now, the inode field is 
 * unused.
 *
 * Why did we do something this gross?  Because the sock structure
 * doesn't contain a way to add arbitrary private data and because
 * af_unix contains most of the fields we want, and because it
 * will always be there no matter what the compilation options
 * are (at least for 2.0).
 */
typedef struct sock panNodeAppLinkSocket;

#define panLOCK if (0)
#define panUNLOCK if (0)

/* State information for the link to a particular application.
 * The NodeAppLink structure is also assumed to be a reference-only 
 * software segment of type panSwseg_TYPE_APPLINK so that it can be
 * handled by code objects.  
 *
 * The applink is referenced by the socket (which sets the 
 * refcount to one, initially).
 *
 * We're currently using AF_PAN sockets for communication.
 * */
typedef struct _panNodeAppLink {
  panSwseg  sws;		/* Used for reference counting and for
				 * passing this object to capsules. 
				 * Must be of type panSwseg_TYPE_APPLINK. */
  panNodeAppLinkSocket *sk;	/* The socket (it points to us too!).
				 * When it goes away, it unreferences
				 * us and sets this to NULL, which may
				 * cause this swseg to be freed.  */
} panNodeAppLink;

#define panNodeAppLinkSocket_getAppLink(sk) \
        ((panNodeAppLink *)((sk)->protinfo.af_unix.inode))
#define panNodeAppLinkSocket_setAppLink(sk, al) \
        (sk)->protinfo.af_unix.inode = (struct inode *)(al);


extern int
panNodeAppLink_deliverData(panNodeAppLink *alink, panStream *datastart, 
			    int length);

extern int 
panNodeAppLink_isConnected(panNodeAppLink *applink);

extern int
panNodeAppLink_proto_init(struct net_proto *pro);

extern void 
panNodeAppLink_proto_remove(void);


#endif /* NODE_APPIF_H_INC */
