/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/linux2/knode.c,v 1.3 1997/12/18 02:33:30 nygren Exp $
 *
 * FILE:    knode.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 *
 */

#define __INCLUDE_KERNEL_VERSION__   /* In this one file ONLY */
#include <common/pan-copyright.h>
#include <common/sysconfig.h>	     /* Includes linux/module.h and other kernel headers */
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <net/ip.h>
#include <net/protocol.h>

#include "app_if.h"
#include "net_if.h"
#include "kpand.h"
#include "knode_stubs.h"


/* Built-in code objects */
extern const unsigned char _panCodeobjBinary_dynamicload_po[];
extern const unsigned      _panCodeobjBinary_dynamicload_po_size;
extern const unsigned char _panCodeobjBinary_insertco_po[];
extern const unsigned      _panCodeobjBinary_insertco_po_size;

jmp_buf panNode_Fail_Jump;
int     panNode_Fail_JumpValid=0;


void
panNode_jump_test(int x)
{
  if (panNode_Fail_JumpValid)
    panJump_longjmp(panNode_Fail_Jump, x);
  panPrint_error("panNode_jump_test: got to end!\n");
}


int
init_module(void)
{
  panCodeobj *dynamicload_co, *insertco_co;
  panNodeAddr _panNodeAddr_LocalAddr, 
    *panNodeAddr_LocalAddr=&_panNodeAddr_LocalAddr;
  void *testmalloc;
  int retval;

  panPrint_msg("panKNode: init_module called.  whee!!!\n");

  register_symtab(0);		/* Don't put any of our symbols into
				 * the kernel's symbol table for now... */

  if (!panNode_init()) {
    panPrint_error("panKNode: couldn't initialize node!\n");
    return -ENODEV;		/* Maybe need a better return value? */
  }
  panPrint_debug("panKNode: panNode_init completed\n");

  testmalloc = panMemory_alloc(500);
  panPrint_debug("panKNode: allocated memory\n");
  panMemory_free(testmalloc);
  panPrint_debug("panKNode: freed memory\n");
    
  panNode_Fail_JumpValid = 1;
  retval = panJump_setjmp(panNode_Fail_Jump);
  if (0 == retval) {
    panPrint_debug("panKNode: going to try to do a jump!\n");
    panNode_jump_test(42);
  } else {
    panPrint_debug("panKNode: jump returned with value %d!\n", retval);
  }

  /* Set for safety from failures while initializing */
  panNode_Fail_JumpValid = 1;
  if (panJump_setjmp(panNode_Fail_Jump)) {
    panPrint_debug("panKNode: fatal error!  exiting init_module!\n");
    return -ENODEV;
  }
  
    /*memset(panNodeAddr_LocalAddr, 0, sizeof(panNodeAddr));*/

  panPrint_debug("panKNode: "
		 "loading panCodeobjBinary_dynamicload_po\n");
  dynamicload_co = 
    panCodeobj_loadFromMemory((char*)_panCodeobjBinary_dynamicload_po,
			      _panCodeobjBinary_dynamicload_po_size,
			      panNode_CodeCache, panNodeAddr_LocalAddr);
  if (!dynamicload_co) {
    panPrint_error("panKNode: couldn't load dynamicload\n");
  }
  /* TODO: change the name of this variable! */
  panCodeobj_DynamicLoadCodeobj = dynamicload_co;

#if 0
  panPrint_debug("panKNode: returning early\n");
  return 0;
#endif

  panPrint_debug("panKNode: "
		 "loading panCodeobjBinary_insertco_po\n");
  insertco_co = panCodeobj_loadFromMemory((char*)_panCodeobjBinary_insertco_po,
			    _panCodeobjBinary_insertco_po_size,
			    panNode_CodeCache, panNodeAddr_LocalAddr);
  if (!insertco_co) {
    panPrint_error("panKNode: couldn't load insertco\n");
  }

  panPrint_debug("panKNode: registering node applink handler.\n");


  /* Initialize the network interface */
  if (retval = panNodeNetIF_init()) {
    panPrint_error("panKNode: network interface initialization failed!\n");
    return retval;
  }

  /* Start up the kpand kernel thread... */
  panKPanD_startup();

  /* Register the node applink protocol handler.  
   * This should allow us to start receiving connections from applications.
   */
  if (retval = panNodeAppLink_proto_init(NULL)) {
    panPrint_error("panKNode: applink  initialization failed!\n");
    return retval;
  }

  panPrint_debug("panKNode: done with init_module.\n");

  return 0;
}

void
cleanup_module(void)
{
  panNodeAppLink_proto_remove();
  panNodeNetIF_cleanup();
  panPrint_msg("panKNode: unloaded.\n");
}



