/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/linux2/knode_util.c,v 1.1 1997/12/10 06:58:34 nygren Exp $
 *
 * FILE:    knode_util.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Implements commonly used functions for allocating memory, etc.
 * Most of these functions simply translate calls to kernel calls
 * with some error checking.
 * 
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>

extern jmp_buf panNode_Fail_Jump;
extern int panNode_Fail_JumpValid;

/* Uses kmalloc to try and allocate memory.  
 * Be aware that this may result in weird-sized blocks of
 * memory.  We need to rework things to deal better with
 * allocating memory within network code as well!
 */
void *
panMemory_alloc(size_t size)
{
  char *mem;
  mem = kmalloc(size, GFP_ATOMIC);
  if (mem) return mem;
  panPrint_warn("panMemory_alloc: couldn't kmalloc %d using GFP_ATOMIC."
		"  Trying again.\n", size);
  mem = kmalloc(size, GFP_KERNEL);
  if (mem) return mem;
  panPrint_error("panMemory_alloc: couldn't kmalloc %d using GFP_KERNEL.\n",
		 size);
  if (panNode_Fail_JumpValid)
    panJump_longjmp(panNode_Fail_Jump, -1);
  return NULL;
}

void *
panMemory_calloc(size_t nmemb, size_t size)
{
  char *mem;
  mem=panMemory_alloc(nmemb*size);
  if (!mem) return NULL;
  memset(mem, 0, nmemb*size);
  return mem;
}

void
panMemory_free(void *ptr)
{
  if (ptr) kfree(ptr);
  else 
    panPrint_error("panMemory_free: attempt to free NULL ptr\n");   
}


void *
panKNodeUtil_memcpy(void *dest, const void *src, size_t n)
{
  return memcpy(dest, src, n);
}

int
panKNodeUtil_memcmp(const void *s1, const void *s2, size_t n)
{
  return memcmp(s1, s2, n);
}

void *
panKNodeUtil_memset(void *s, int c, size_t n)
{
  return memset(s, c, n);
}


unsigned long int
panKNodeUtil_htonl(unsigned long int hostlong)
{
  return htonl(hostlong);
}

unsigned short int
panKNodeUtil_htons(unsigned short int hostshort)
{
  return htons(hostshort);
}

unsigned long int
panKNodeUtil_ntohl(unsigned long int netlong)
{
  return ntohl(netlong);
}

unsigned short int
panKNodeUtil_ntohs(unsigned short int netshort)
{
  return ntohs(netshort);
}
