/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/linux2/skbuffswseg.c,v 1.2 1997/12/18 02:33:34 nygren Exp $
 *
 * FILE:    skbuffswseg.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for skbuff software segment data structures and routines.
 * See skbuffswseg.h for more documentation.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <linux/kernel.h>
#include <asm/atomic.h>
#include "skbuffswseg.h"


/* Finalizes a panSwseg_TYPE_SKBUFF swseg by freeing (decrimenting the count
 * on) the underlying swseg.  Also frees the swseg wrapper.
 * 
 */
void
panSwsegSkbuff_finalize(panSwseg *sws)
{
  panSwsegSkbuff *swsskb;

  if (!sws) return;

  swsskb = (panSwsegSkbuff *)sws;
  if (sws->type != panSwseg_TYPE_SKBUFF)
    { panPrint_error("panSwsegSkbuff_finalize called with segment of type %d\n", sws->type); return; }
  if (sws->next) panSwseg_unref(sws->next);
  if (swsskb->skb)
    {
      int needfree=0;
      panPrint_debug("panSwsegSkbuff_finalize: calling kfree_skb on skbuff "
		     "%p (count=%d, lock=%d, onlist=%p)\n",
		     swsskb->skb, swsskb->skb->count, swsskb->skb->lock,
		     swsskb->skb->list);
      /* Free it (ie, decriment count and free as appropriate) */
      {
	unsigned long flags;
	save_flags(flags);
	cli();
	if (swsskb->skb->count>1)
	  swsskb->skb->count--;
	else
	  needfree=1;
	restore_flags(flags);
      }
      if (needfree)
	kfree_skb(swsskb->skb, swsskb->creditto);      
    }
  panMemory_free(sws);
}


/* Creates a swseg with a refcount of 1 that contains an skbuff.
 * Does not incriment the skbuff's count.  The swseg contains
 * the skbuff's active region.
 */
panSwseg *
panSwsegSkbuff_create_fromSkbuff(struct sk_buff *skb)
{
  panSwsegSkbuff *newsws;
  
  if (!skb)
    { panPrint_error("panSwsegSkbuff_create_fromSkbuff: passed null skbuff\n");
    return NULL; }
  newsws = panMemory_allocStruct(panSwsegSkbuff);
  if (!newsws)
    { panPrint_error("panSwsegSkbuff_create_fromSkbuff: malloc returned null\n");
    return NULL; }

  newsws->sws.length = skb->len;
  newsws->sws.data   = skb->data;
  newsws->sws.type = panSwseg_TYPE_SKBUFF;
  newsws->sws.numref = 1;
  newsws->sws.next = NULL;
  newsws->sws.finalize = panSwsegSkbuff_finalize;
  newsws->skb = skb;

  if (skb->sk)
    panPrint_warn("panSwsegSkbuff_create_fromSkbuff: skb has sk set\n");

#if PAN_DEBUG
  if (ADDR_NOT_ALIGNED_P(newsws->sws.length)
      || ADDR_NOT_ALIGNED_P(newsws->sws.data))
    panPrint_warn("panSwsegSkbuff_create_fromSkbuff: length (%d) or data (%p) not aligned properly\n", newsws->sws.length, newsws->sws.data);
#endif
  
  return (panSwseg*)newsws;
}


/* Creates a swseg with a refcount of 1 that contains a new skbuff.
 * Does not incriment the skbuff's count.  The swseg contains
 * the skbuff's active region with the reserved head and tail room.
 */
panSwseg *
panSwsegSkbuff_create(unsigned size, unsigned headroom, unsigned tailroom)
{
  panSwsegSkbuff *newsws;
  struct sk_buff *skb;

  skb = alloc_skb(size+headroom+tailroom, GFP_ATOMIC);
  if (!skb)
    { panPrint_error("panSwsegSkbuff_create: creation of skbuff failed\n");
    return NULL; }
  newsws = panMemory_allocStruct(panSwsegSkbuff);
  if (!newsws)
    { panPrint_error("panSwsegSkbuff_create: malloc returned null\n");
    return NULL; }

  skb->free = 1;	                    /* how to free */
  if (headroom) skb_reserve(skb, headroom); /* reserve some headroom */
  newsws->sws.data = skb_put(skb, size);    /* put a region of size following
					     * the headroom and make
					     * the data pointer point to it. */

  newsws->sws.length = size;
  newsws->sws.type = panSwseg_TYPE_SKBUFF;
  newsws->sws.numref = 1;
  newsws->sws.next = NULL;
  newsws->sws.finalize = panSwsegSkbuff_finalize;
  newsws->skb = skb;

#if PAN_DEBUG
  if (ADDR_NOT_ALIGNED_P(newsws->sws.length)
      || ADDR_NOT_ALIGNED_P(newsws->sws.data))
    panPrint_warn("panSwsegSkbuff_create: length (%d) or data (%p) not aligned properly\n", newsws->sws.length, newsws->sws.data);
#endif

  return (panSwseg*)newsws;
}


