/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/app_if.h,v 1.4 1997/12/13 05:14:33 nygren Exp $
 *
 * FILE:    app_if.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides routines for communicating with applications connected
 * to a PAN node over UNIX domain sockets.
 *
 */

#ifndef NODE_APPIF_H_INC
#define NODE_APPIF_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <common/sysconfig.h>
#include <node/nodecodeobj.h>
#include <common/nodeaddr.h>
#include <common/conventions.h>

#include <sys/un.h>


/* State information for the link to a particular application.
 * The NodeAppLink structure is also assumed to be a reference-only 
 * software segment of type panSwseg_TYPE_APPLINK so that it can be
 * handled by code objects.  The reference count may drop to zero at
 * some points, but it will only be removed by the finalizer when
 * it is also disconnected.
 *
 * We're currently using UNIX domain socket streams where
 * a 4-byte integer capsule length is sent, followed by the capsule
 * itself (this repeats).
 */
typedef struct _panNodeAppLink {
  panSwseg  sws;		/* Used for reference counting and for
				 * passing this object to capsules. 
				 * Must be of type panSwseg_TYPE_APPLINK. */
  struct _panNodeAppLink *prev, *next; /* For chaining together in
					* appif->applinks list */
  int sock;			/* The UNIX domain socket bound
				 * to the application */
  int status;			/* Is this link connected? */
#define panNodeAppLink_STATUS_CONNECTED    0
#define panNodeAppLink_STATUS_DISCONNECTED 1
  
  int capsize;   		/* The number of bytes coming through
				 * in the next capsule from the application.
				 * If the next thing to do is to read
				 * another 4-byte capsule length, then has
				 * the value -1.  */
  struct sockaddr_un appsockaddr; /* Contains the address (filename) of the 
				   * application connected at the
				   * other name of the socket */
  struct _panNodeAppIF *appif;	/* The application interface */
} panNodeAppLink;




/* Overall state for the node's interface applications.
 * Also contains info about the port that we're listening
 * for connections from new applications on.
 */
typedef struct _panNodeAppIF {
  struct sockaddr_un listensockaddr; /* Contains the filename of the 
				      * socket we're listening at 
				      * for connections from applications. */
  int                listensock; /* The socket we're listing at for 
				  * connections from applications.  */
  panNodeAppLink   *applinks;	/* A list of all applinks */
} panNodeAppIF;


extern panNodeAppIF *
panNodeAppIF_create(u_int16_t portnum);

extern panNodeAppLink *
panNodeAppIF_accept(panNodeAppIF *appif);

extern void
panNodeAppLink_recvAndEval(panNodeAppLink *alink);

extern int
panNodeAppLink_deliverData(panNodeAppLink *alink, panStream *datastart, 
			    int length);

extern int 
panNodeAppLink_isConnected(panNodeAppLink *applink);

extern void
panNodeAppLink_disconnect(panNodeAppLink *alink);

extern void
panNodeAppLink_destroy(panNodeAppLink *alink);



#endif /* NODE_APPIF_H_INC */
