/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/net_if.h,v 1.4 1998/01/23 07:43:30 nygren Exp $
 *
 * FILE:    net_if.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides a network interface to a pan UNode.  Contains routines
 * for creating network sockets and sending/receiving capsules.
 *
 */

#ifndef NODE_NETIF_H_INC
#define NODE_NETIF_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <common/sysconfig.h>
#include <node/nodecodeobj.h>
#include <common/nodeaddr.h>

#include <netinet/in.h>

#include "routing.h"

/* Number of input buffers (software segments) */
#define panUNodeNetIF_INBUF_COUNT 16

/* Size in bytes of each input buffer */
#define panUNodeNetIF_INBUF_SIZE  512

/* The maximum size that a capsule may be in the existing system 
 * (due to the size of input buffers). */
#define panUNodeNetIF_MAX_CAPSULE_SIZE \
             (panUNodeNetIF_INBUF_SIZE*panUNodeNetIF_INBUF_COUNT)

/* Maximum number of swsegs in a capsule being sent */
#define panUNodeNetIF_MAX_SEND_CAPSULE_SWSEGS 64

/* State information for a UNode's network interface. */
typedef struct _panUNodeNetIF {
  panRouteTable *rtab;		/* Routing table */
  panNodeAddr   *localaddr;	/* Address of local network node */
  int            netsock;	/* UDP network socket */

  /* We currently use a fixed number of fixed size input buffers
   * that data is filled into on receive.  It may make sense to
   * have variable sized input buffers (increasing in size by
   * a factor of two or something) in the future.
   * By reusing the unused swsegs and iovec entries, we can reduce the 
   * amount of work we need to do.
   * 
   * Any routines that consume buffers or rcvvec slots should
   * make sure to replenish them all before exiting.
   */
  panSwseg      *rcvbufs[panUNodeNetIF_INBUF_COUNT];  /* A collection of input
						       * buffers to use for
						       * receiving capsules. */
  struct iovec   rcvvecs[panUNodeNetIF_INBUF_COUNT]; /* A collection of iovecs
						      * to use for receiveing
						      * capsules (point to 
						      * bufs in rcvbufs). */
} panUNodeNetIF;


extern panUNodeNetIF *panNode_NetIF;

extern panUNodeNetIF *
panUNodeNetIF_create(panRouteTable *rtab, panNodeAddr *localaddr);

extern void
panUNodeNetIF_recvAndEval(panUNodeNetIF *netif);

extern int
panUNodeNetIF_sendCapsule(panUNodeNetIF *netif,
			  panNodeAddr *destaddr, panStream *capstart, 
			  int length);

extern u_int16_t
panUNodeNetIF_getPortNum(panUNodeNetIF *netif);


#endif /* NODE_NETIF_H_INC */
