/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/routing.c,v 1.5 1997/12/13 05:14:33 nygren Exp $
 *
 * FILE:    routing.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 *
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/nodeaddr.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include "routing.h"


/* The address of this node */
panNodeAddr *panNodeAddr_LocalAddr = NULL;

/* Then routing table for this node */
panRouteTable panRouteTable_RTab = { 0, NULL };



/* Compares two panRouteEntry's for sorting purposes.
 * Returns 0 if they're equal, -1 if una1<una2, and 1 if una1>una2.
 */
static __inline__ int
panRouteEntry_compare(panRouteEntry *re1, panRouteEntry *re2)
{
  return panNodeAddr_compare(&(re1->dest), &(re2->dest));
}


/* Stores the node's local nodeaddr into na */
void
panNodeAddr_getLocalAddr(panNodeAddr *na)
{
  if (na&&panNodeAddr_getLocalAddr)
    panNodeAddr_dup(na, panNodeAddr_LocalAddr);
}


/* Looks through the routing table for an entry with
 * dest==una and returns it.  Assumes that rt is sorted.  
 */
panRouteEntry *
panRouteTable_lookup(panRouteTable *rt, panNodeAddr *una)
{
  int p, q, m;			/* Bounds for binary search and midpoint */
  int cmp;

#if 0
  if (!rt->numentries) return NULL;
  p=0; q=rt->numentries-1;
  while (p <= q)
    {
      m = (p+q)/2;
      cmp = panNodeAddr_compare(&(rt->rtab[m].dest), una);
      panPrint_debug("looking up route: p=%d m=%d q=%d cmp=%d\n",p,m,q,cmp);
      if (0==cmp) 
	return &(rt->rtab[m]);
      else if (cmp>0) /* rtab[m] < una */
	p = m+1;
      else            /* rtab[m] > una */
	q = m-1;
    }
  return NULL;
#else
  for (q=0; q<rt->numentries; q++)
    {
      if (0==panNodeAddr_compare(&(rt->rtab[q].dest), una))
	{
	  panPrint_debug("panRouteTable_lookup: going through %d to reach ",
			 q);
	  panNodeAddr_print(&(rt->rtab[q].dest));
	  panPrint_debug("\n");
	  return &(rt->rtab[q]);
	}
    }
  return NULL;
#endif
}


/* Prints out the routing table for this node. */
void
panRouteTable_print(panRouteTable *rt)
{
  int i;
  unsigned char *addr;
  
  panPrint_msg("Routing table:\n\tdest\t\t\t\tnexthop\n");
  for (i=0; i<rt->numentries; i++)
    {    
      panPrint_msg("\t");
      panNodeAddr_print(&(rt->rtab[i].dest));
      addr = (unsigned char*)&(rt->rtab[i].nexthop.sin_addr);
      panPrint_msg("\t=>\t%d.%d.%d.%d:%d\n",addr[0],addr[1],addr[2],addr[3],
		   ntohs(rt->rtab[i].nexthop.sin_port));
    }
}

