/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/routing.h,v 1.4 1997/12/01 22:16:16 nygren Exp $
 *
 * FILE:    routing.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Routines for maintaining routing tables for a PAN UNode.
 * A routing table is a mapping from PAN UNode addresses to
 * sockaddr_in structures.
 *
 * A PAN UNode address is a 4-byte IP address followed by
 * a 2-byte port number.  Both are in network byte order.
 *
 */

#ifndef NODE_ROUTING_H_INC
#define NODE_ROUTING_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <common/sysconfig.h>
#include <node/nodecodeobj.h>
#include <common/nodeaddr.h>

#include <netinet/in.h>

#include <siod/siod.h>


/* An internal representation for a PAN node address.
 * In capsules, all PAN addresses are 128 bits, so 80 bits of
 * unused padding follow when on-the-wire. 
 * For UNodes, the panNodeAddr contains a 
 * 4-byte IP addr followed by 2-byte portnum followed 
 by 10 bytes of padding.
 */

/* An entry in a routing table for a PAN node. */
typedef struct _panRouteEntry {
  panNodeAddr  dest;		/* The destination we're trying to get to.
				 * Must be the first field, 
				 * so packed at front. */
  struct sockaddr_in   nexthop;	/* Where to route packets going to dest */
} panRouteEntry;


/* A routing table contains mappings from destinations to nexthops. 
 * Creation takes O(n lg n) average time and lookups take O(lg n) time
 * for this implementation. */
typedef struct _panRouteTable {
  int            numentries;	/* The size of the routing table */
  panRouteEntry *rtab;		/* The routing table.  This is an
				 * array of panRouteEntry fields.
				 * It must be sorted by dest fields.
				 * This way, we can do a binary search
				 * in O(lg n) time.  A hash table might
				 * be even faster, but that's for some
				 * future time. */
} panRouteTable;



/* The address of this node */
extern panNodeAddr *panNodeAddr_LocalAddr;

/* Then routing table for this node */
extern panRouteTable panRouteTable_RTab;



extern unsigned
panNodeAddr_read(panNodeAddr *una, panStream *ps);

extern unsigned
panNodeAddr_write(panStream *ps, panNodeAddr *una);

extern void
panNodeAddr_init(panNodeAddr *una, u_int32_t addr, u_int16_t port);

extern panRouteEntry *
panRouteTable_lookup(panRouteTable *rt, panNodeAddr *una);

extern void
panRouteTable_print(panRouteTable *rt);

extern LISP 
panUNodeCfg_setNodeAddr(LISP nodeaddr);

extern LISP 
panUNodeCfg_setRouteTable(LISP nodeaddr);


#endif /* NODE_ROUTING_H_INC */
