/* $XFree86: xc/programs/Xserver/hw/xfree86/loader/loader.c,v 1.12 1997/04/08 10:13:48 hohndel Exp $ */




/*
 *
 * Copyright 1995,96 by Metro Link, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Metro Link, Inc. not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Metro Link, Inc. makes no
 * representations about the suitability of this software for any purpose.
 *  It is provided "as is" without express or implied warranty.
 *
 * METRO LINK, INC. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL METRO LINK, INC. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/symtab.h>

#include "loader.h"
#include "ar.h"
#include "elf.h"
#include "coff.h"

#define AOUT_AND_ELF_ONLY

/*
#define DEBUG
#define DEBUGAR
#define DEBUGLIST
#define DEBUGMEM
*/

int check_unresolved_sema = 0;


void
dummy()
{
  return;
}

int 
dummy0(dum)
     int dum;
{
  return 0;
}

/*
 * Array containing entry points for different formats.
 */

static panBinLoaderFuncs loaderfuncs[] = {
#ifndef AOUT_AND_ELF_ONLY
  /* LD_ARCHIVE */
  {ARCHIVELoadModule,dummy,(int(*)())dummy0,dummy},
  /* LD_ELFOBJECT */
  {ELFLoadModule,ELFResolveSymbols,ELFCheckForUnresolved,ELFUnloadModule},
  /* LD_COFFOBJECT */
  {COFF2LoadModule,COFF2ResolveSymbols,COFF2CheckForUnresolved,COFF2UnloadModule},
  /* LD_XCOFFOBJECT */
  {COFF2LoadModule,COFF2ResolveSymbols,COFF2CheckForUnresolved,COFF2UnloadModule},
  /* LD_AOUTOBJECT */
  {AOUTLoadModule,AOUTResolveSymbols,AOUTCheckForUnresolved,AOUTUnloadModule},
#else
  /* Place holder for LD_ARCHIVE*/
  {AOUTLoadModule,AOUTResolveSymbols,AOUTCheckForUnresolved,AOUTUnloadModule},
  /* LD_ELFOBJECT */
  {ELFLoadModule,ELFResolveSymbols,ELFCheckForUnresolved,ELFUnloadModule},
  /* Place holder for LD_COFFOBJECT*/
  {AOUTLoadModule,AOUTResolveSymbols,AOUTCheckForUnresolved,AOUTUnloadModule},
  /* Place holder for LD_XCOFFOBJECT*/
  {AOUTLoadModule,AOUTResolveSymbols,AOUTCheckForUnresolved,AOUTUnloadModule},
  /* LD_AOUTOBJECT */
  {AOUTLoadModule,AOUTResolveSymbols,AOUTCheckForUnresolved,AOUTUnloadModule},
#endif /* AOUT_AND_ELF_ONLY */
#ifdef DLOPEN_SUPPORT
  /* LD_DLOPEN */
  {DLLoadModule,DLResolveSymbols,DLCheckForUnresolved,DLUnloadModule},
#endif
};

int	numloaders=sizeof(loaderfuncs)/sizeof(panBinLoaderFuncs);


/*
 * Determine what type of object is being loaded.
 * Looks at the bincode field of the code object.
 */
static int
panBinLoader_getCodeType(panCodeobj *codeobj)
{
  unsigned char	*buf; /* long enough for the largest magic type */

  if (codeobj->bincode_sz < 10)
    return -1;

  buf = codeobj->bincode;

  if (strncmp((char*)buf,ARMAG,SARMAG) == 0 ) {
    return LD_ARCHIVE;
  }

  if (strncmp((char*)buf,ELFMAG,SELFMAG) == 0 ) {
    return LD_ELFOBJECT;
  }

  if (buf[0] == 0x4c && buf[1] == 0x01 ) {
    /* I386MAGIC */
    return LD_COFFOBJECT;
  }
  if (buf[0] == 0x01 && buf[1] == 0xdf ) {
    /* XCOFFMAGIC */
    return LD_COFFOBJECT;
  }
  if (buf[0] == 0x0d && buf[1] == 0x01 ) {
    /* ZCOFFMAGIC (LynxOS) */
    return LD_COFFOBJECT;
  }
  if (buf[0] == 0x00 && buf[1] == 0x86 && buf[2] == 0x01 && buf[3] == 0x07) {
    /* AOUTMAGIC */
    return LD_AOUTOBJECT;
  }
  if (buf[0] == 0x07 && buf[1] == 0x01 && (buf[2] == 0x64 || buf[2] == 0x86)) {
    /* AOUTMAGIC, (Linux OMAGIC, old impure format, also used by OS/2 */
    return LD_AOUTOBJECT;
  }
  return LD_UNKNOWN;
}

/* Used when loading archives */
static unsigned long offsetbias = 0;

/*
 * panBinLoader_getMemRegion returns a pointer to the start of the specified
 * memory region within code object's bincode section.
 * This memory should be treated as read-only!
 */
void *
panBinLoader_getMemRegion(panCodeobj *codeobj, unsigned long offset,
			  int size, char *label)
{
  u_int8_t *ptr;

#ifdef DEBUGMEM
  panPrint_debug("panBinLoader_getMemRegion(%u(%u),%d,%s)",offset,offsetbias,size,label);
#endif

  if (!codeobj || !codeobj->bincode
      || (codeobj->bincode_sz < offset+size))
    { panPrint_warn("panBinLoader_getMemRegion: short bincode section when reading %s\n", label); return NULL; }

  ptr = codeobj->bincode + offset;

#ifdef DEBUGMEM
  panPrint_debug("=%x\n",ptr);
#endif

  return (void *)ptr;
}

/*
 * Releases a region of memory returned by panBinLoader_getMemRegion.
 */
void
panBinLoader_releaseMemRegion(void *addr, int size)
{
  return;
}


/* panBinLoader_copyRegionToBuf copies memory into a buffer from the
 * start of the specified memory region within code object's bincode
 * section.
 */
int 
panBinLoader_copyRegionToBuf(panCodeobj *codeobj, unsigned int offset,
				 void *buf, int size)
{
  char *mem;

  mem = panBinLoader_getMemRegion(codeobj, offset, size, "copyRegionToBuf");
  if (!mem || !buf) return -1;
  memcpy(buf, mem, size);
  panBinLoader_releaseMemRegion(mem, size);

  return size;
}

/* 
 * panBinLoader_handleUnresolved decides what to do with an unresolved
 * symbol. Right now, it will print out the unresolved symbol.
 * Returns 0 if the unresolved symbol should be ignored.
 * Otherwise it returns 1, indicating failure.
 */

int
panBinLoader_handleUnresolved(panCodeobj *codeobj, char *symbol)
{
  int fatalsym = 0;

  panPrint_warn("panBinLoader_handleUnresolved: couldn't resolve symbol <%s>"
		" while loading code object <%s>\n",
		symbol, codeobj->symname);

  fatalsym = 1;
  return(fatalsym);
}

/* Loads binary code into the panCodeobj codeobj from 
 * memory pointed to by the bincode field of codeobj.
 * Uses the code object's symtab and its dependencies'
 * symtabs and the global symtab to resolve unresolved symbols.
 * Returns 1 on success.  Returns 0 if an error occurs during the load. 
 * Does not set fn_* in the code object (this should be done
 * by the caller).  */
int 
panBinLoader_load(panCodeobj *codeobj)
{
  int codetype;
  panBinLoaderRec *blrec; 
  
  assert(codeobj);
  if (!codeobj->bincode) 
    { panPrint_warn("panBinLoader_load: codeobj <%s> with no bincode\n",
		    codeobj->symname); return 0; }
  if (codeobj->status > 0)
    { panPrint_warn("panBinLoader_load: dependencies unavailable for <%s>\n",
		    codeobj->symname); return 0; }

  if ((codetype=panBinLoader_getCodeType(codeobj)) < 0)
    { panPrint_warn("panBinLoader_load: unknown bincode type for <%s>\n", 
		    codeobj->symname); return 0; }
  
  if (!(blrec=(panBinLoaderRec*)panMemory_alloc(sizeof(panBinLoaderRec))))
    { panPrint_error("panBinLoader_load: couldn't alloc blrec\n"); return 0; }
  
  blrec->funcs   = &(loaderfuncs[codetype]);
  blrec->codeobj = codeobj;

  panPrint_debug("panBinLoader_load: About to call loadmodule for "
		 "<%s> of type %d\n", codeobj->symname, codetype);
  if (NULL == (blrec->private
	       = blrec->funcs->LoadModule(codetype, codeobj->symname, 
					   codeobj)))
    { 
      panPrint_warn("panBinLoader_load: failed to load codeobj <%s>\n",
		     codeobj->symname);
      panMemory_free(blrec); return 0; 
    }
  panPrint_debug("panBinLoader_load: returned from loadmodule\n");

  codeobj->loader_priv = (void*)blrec;

  blrec->funcs->ResolveSymbols(codeobj);

  if (blrec->funcs->CheckForUnresolved(codeobj))
    {
      panPrint_warn("panBinLoader_load: unresolved symbols in <%s>\n",
		     codeobj->symname);
      panBinLoader_unload(codeobj);
      return 0; 
    }

  return 1;
}


/* Unloads the bincode from codeobj */
void
panBinLoader_unload(panCodeobj *codeobj)
{
  panBinLoaderRec *blrec;

  if (!codeobj || !codeobj->loader_priv) return;
  
  panPrint_debug("panBinLoader_unload: unloading <%s>\n", codeobj->symname);
  blrec = (panBinLoaderRec*)codeobj->loader_priv;
  blrec->funcs->LoaderUnload(blrec->private);
  panMemory_free(blrec);
}







#ifndef AOUT_AND_ELF_ONLY
/*
 * Handle an archive.
 */
void *
ARCHIVELoadModule(modtype,modname,handle,arfd)
     int	modtype; /* Always LD_ARCHIVE */
     char	*modname;
     int	handle;
     int	arfd;
{
  loaderPtr tmp ;
  unsigned char	magic[SARMAG];
  struct ar_hdr	hdr;
  unsigned int	size;
  unsigned int	offset = 0;
  int	arnamesize, modnamesize;
  char	*slash;

  if( modtype != LD_ARCHIVE ) {
    ErrorF( "ARCHIVELoadModule(): modtype != ARCHIVE\n" );
    return NULL;
  }

  arnamesize=strlen(modname);

  read(arfd,magic,SARMAG);

  if(strncmp((const char *)magic,ARMAG,SARMAG) != 0 ) {
    ErrorF("ARCHIVELoadModule: wrong magic!!\n" );
    return NULL;
  }

  /* Skip the symbol table */
  read(arfd,&hdr,sizeof(struct ar_hdr));

  if( hdr.ar_name[0] == '/' || strncmp(hdr.ar_name, "__.SYMDEF", 9) == 0) {
    /* If the file name is NULL, then it is a symbol table */
    sscanf(hdr.ar_size,"%d",&size);
#ifdef DEBUGAR
    ErrorF("Member '%16.16s', size %d, offset %d\n",
	   hdr.ar_name, size, offset );
    ErrorF("Symbol table size %d\n", size );
#endif
    offset=lseek(arfd,size,SEEK_CUR);
    if( offset&0x1 ) /* odd value */
      offset=lseek(arfd,1,SEEK_CUR); /* make it an even boundary */
  }
  else	{
    /* No symbol table - reset the file position to the first file */
    offset=lseek(arfd,SARMAG,SEEK_SET);
  }

  /* Skip the string table */
  read(arfd,&hdr,sizeof(struct ar_hdr));

  if( hdr.ar_name[0] == '/' && hdr.ar_name[1] == '/') {
    /* If the file name is '/', then it is a string table */
    sscanf(hdr.ar_size,"%d",&size);
#ifdef DEBUGAR
    ErrorF("Member '%16.16s', size %d, offset %d\n",
	   hdr.ar_name, size, offset );
    ErrorF("String table size %d\n", size );
#endif
    offset=lseek(arfd,size,SEEK_CUR);
    if( offset&0x3 ) /* needs to be long word aligned */
      offset=lseek(arfd,4-(offset&0x3),SEEK_CUR);
  }
  else	{
    /* No string table - reset the file position to the first file */
    offset=lseek(arfd,offset,SEEK_SET);
  }

  while( read(arfd,&hdr,sizeof(struct ar_hdr)) ) {

    sscanf(hdr.ar_size,"%d",&size);
    offset=lseek(arfd,0,SEEK_CUR);

#ifdef DEBUGAR
    panPrint_debug("Member '%16.16s', size %d, offset %d\n",
	   hdr.ar_name, size, offset );
#endif

    slash=strchr(hdr.ar_name,'/');
    if (slash == NULL) {
      /* BSD format without trailing slash */
      slash = strchr(hdr.ar_name,' ');
    } 
    /* XXX lots to do with long name in the form #1/ */
    /* SM: Make sure we do not overwrite other parts of struct */
        
    if((slash - hdr.ar_name) > sizeof(hdr.ar_name)) 
      slash = hdr.ar_name + sizeof(hdr.ar_name) -1;
    *slash='\000';

    if( (modtype=_GetModuleType(arfd,offset)) < 0 ) {
      ErrorF( "%s is an unrecognized module type\n", hdr.ar_name ) ;
      offsetbias=0;
      return NULL;
    }

    tmp=_LoaderListPush();

    tmp->handle = handle;
    tmp->funcs=funcs[modtype];
    modnamesize=strlen(hdr.ar_name);
    tmp->name=(char *)panMemory_alloc(arnamesize+modnamesize+2 );
    strcpy(tmp->name,modname);
    strcat(tmp->name,":");
    strcat(tmp->name,hdr.ar_name);

    offsetbias=offset;

    if( (tmp->private=funcs[modtype].LoadModule(modtype,tmp->name,handle,
						arfd)) == NULL ) {
      ErrorF( "Failed to load %s\n", hdr.ar_name ) ;
      offsetbias=0;
      return NULL;
    }

    offset=lseek(arfd,offset+size,SEEK_SET);
    if( offset&0x1 ) /* odd value */
      lseek(arfd,1,SEEK_CUR); /* make it an even boundary */
  }
  offsetbias=0;

  return tmp->private;
}

#endif /* AOUT_AND_ELF_ONLY */



int
panBinLoader_defaultFunc(void)
{
  panPrint_error("\n\n\t*** This should not happen!\n\t*** An unresolved function was called!\n\n");
  return 0;
}

