/* $Header: /home/dl2/nygren/cvs/pan/src/node/codecache.c,v 1.4 1997/12/18 02:33:23 nygren Exp $
 *
 * FILE:    codecache.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Implementation of routines for caching code objects in a PAN node.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>

/*
 *  Creates a new, empty code cache and returns a pointer to it
 */
panCodeCache *
panCodeCache_create(void)
{
  panCodeCache *cc;
  
  cc = panMemory_calloc(1,sizeof(panCodeCache));
  if (!cc) 
    { panPrint_error("alloc failed in panCodeCache_create\n"); return NULL; }
  cc->codelist = panCodeobjList_create();
  if (!cc->codelist)
    { panPrint_error("failed to create CodeobjList in panCodeCache_create\n"); 
      panMemory_free(cc); return NULL; }
  return cc;  
}


/* Inserts a code object into the code cache */
void
panCodeCache_insert(panCodeCache *cache, panCodeobj *co)
{
  unsigned hash;

  assert(cache&&co);
  hash = panCodeCache_HASH_GENERATE(&(co->name));

  panCodeobjList_addToTail(cache->codelist, co);
  {
    panCritRegion_ENTER();    
    if (!(cache->hashtab[hash]))
      {
	cache->hashtab[hash] = panCodeobjList_create();  
	if (!(cache->hashtab[hash])) {
	  panPrint_error("panCodeCache_insert: failed to create"
			 " CodeobjList\n"); 
	  panCritRegion_LEAVE();
	  return; 
	}
      }
    panCodeobjList_addToTail(cache->hashtab[hash], co);
    panCritRegion_LEAVE();
  }
}

/* Looks up a code object in the code cache and returns it
 * if its there.  Otherwise returns NULL. */
panCodeobj *
panCodeCache_lookup(panCodeCache *cache, panCodeobjName *con)
{
  return panCodeobjList_lookupByName(cache->hashtab[panCodeCache_HASH_GENERATE(con)], 
				     con);
}
