/* $Header: /home/dl2/nygren/cvs/pan/src/node/codecache.h,v 1.2 1997/06/24 21:04:13 nygren Exp $
 *
 * FILE:    codecache.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * The PAN code cache keeps a collection of PAN code objects
 * that can be rapidly looked up.  In the future, it may allow
 * code objects to be managed by keeping track of how frequently
 * they are used.
 *
 */

#ifndef NODE_CODECACHE_H_INC
#define NODE_CODECACHE_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <node/nodecodeobj.h>

/* Number of entries in the code cache.  Must be a power of 2 for
 * panCodeCache_HASH_GENERATE to work */
#define panCodeCache_HASH_ENTRIES 128

/* Takes a code object name ptr and returns the hash table index 
* (a value from 0 to panCodeCache_HASH_ENTRIES-1) */
#define panCodeCache_HASH_GENERATE(con) (((con)->data[1]) & (panCodeCache_HASH_ENTRIES-1))


/* The code cache contains both a list of all loaded
 * code objects (this should be switched to a binary heap at some point)
 * and a hash table.  The hash table contains panCodeobjLists or NULL
 * if nothing is in that bucket.
 */
typedef struct _panCodeCache {
  panCodeobjList *codelist;	/* A list of all loaded code objects */
  panCodeobjList *hashtab[panCodeCache_HASH_ENTRIES];
} panCodeCache;


extern panCodeCache *
panCodeCache_create(void);

extern void
panCodeCache_insert(panCodeCache *cache, panCodeobj *co);

extern panCodeobj *
panCodeCache_lookup(panCodeCache *cache, panCodeobjName *con);



#endif /* NODE_CODECACHE_H_INC */
