/* $Header: /home/dl2/nygren/cvs/pan/src/node/environment.h,v 1.4 1997/12/10 06:58:28 nygren Exp $
 *
 * FILE:    environment.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for capsule execution environments.
 * (An environment contains information about the 
 * capsule currently being evaluated).
 *
 */

#ifndef NODE_ENVIRONMENT_H_INC
#define NODE_ENVIRONMENT_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/symtab.h>
#include <node/nodecodeobj.h>


/* The environment for a capsule being evaluated.
 * In the current system, only a single capsule may
 * be evaluated at a time.  Often, a single environment
 * may be created once and then cleared and reused as needed.
 */
typedef struct _panCapsuleEnv {
  struct _panCapsuleEnv *waitnext; /* Next in a list of capsule environments
				    * waiting to be executed. */
  panSwsegRing   *sring;	/* A ring that keeps track of the
				 * software segments being referenced by
				 * the capsule being evaluated. */
  panCodeobj     *codeobj;	/* The code object for the capsule
				 * being evaluated.  */
  panStream       capstart;	/* The start of the capsule */
  int             caplength;	/* The length of the capsule */
  jmp_buf         exitjump;	/* A jump point for terminating capsule
				 * execution. */
  int             method;	/* The method being executed */
#define panCapsuleEnv_METHOD_INIT     1
#define panCapsuleEnv_METHOD_ACCEPT   2
#define panCapsuleEnv_METHOD_FINALIZE 3
  panNodeAddr     homeaddr;	/* Where to go for referenced code objects */
  panSwseg       *applink;	/* The app that inserted this capsule */
  int             status;	/* Is the capsule executing or not? */
#define panCapsuleEnv_STATUS_DONE    0 /* Not running */
#define panCapsuleEnv_STATUS_RUNNING 1 /* Running */
#define panCapsuleEnv_STATUS_WAITING 2 /* Waiting on availability of code */
} panCapsuleEnv; 


/* A panCapsuleEnvList keeps a list of panCapsuleEnv's.  Each one
 * may only be in a single list at any one time.  */
typedef struct _panCapsuleEnvList {
  panCapsuleEnv *head;
  panCapsuleEnv *tail;  
} panCapsuleEnvList;


extern void
panNode_acceptCapsule(panStream *capdata, int caplength,
		    panSwseg *applink);

extern void
panCodeobj_evalMethod(panCodeobj *co, int method);

extern panCapsuleEnv *
panCapsuleEnv_create(void);

extern void
panCapsuleEnv_destroy(panCapsuleEnv *cenv);

extern void
panCapsuleEnv_evaluate(panCapsuleEnv *cenv);

extern void
panCapsuleEnv_clearAll(panCapsuleEnv *cenv);

extern void
panCapsuleEnv_abort(panCapsuleEnv *cenv, char *msg);



#define panCapsuleEnvList_EMPTY_P(celist) (!((celist) && (celist)->head))

extern panCapsuleEnvList *
panCapsuleEnvList_create(void);

extern void
panCapsuleEnvList_insert(panCapsuleEnvList *celist, panCapsuleEnv *cenv);

extern void
panCapsuleEnvList_evalAll(panCapsuleEnvList *celist);

extern void
panCapsuleEnvList_destroy(panCapsuleEnvList *celist);

extern void
panCapsuleEnvList_transferList(panCapsuleEnvList *cel_dst,
			       panCapsuleEnvList *cel_src);


#endif /* NODE_ENVIRONMENT_H_INC */
