/* $Header: /home/dl2/nygren/cvs/pan/src/node/globalsyms_platindep.c,v 1.10 1998/01/20 23:44:56 nygren Exp $
 *
 * FILE:    globalsyms_platindep.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * List of all globally exported symbols with which to initialize the 
 * symbol table.  These should all be platform-independent (ie, should
 * use the same function regardless of whether we're in the kernel or
 * in user-space).
 *
 * Function names should start with pni (for PAN Node Interface).
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/symtab.h>

#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pninetwork.h>
#include <pni/pnicodeobj.h>
#include <pni/pniapplink.h>
#include <pni/pniutil_hashtab.h>
#include <pni/pnitimestamp.h>

#if 0
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#endif


panSymbol panSymTab_GlobalSymbols_PlatformIndependent[] = {

  /***************** From pni/pniapplink.h *****************/
  panSymbol_MAKEFUNC(pniAppLink_getSender)

  /***************** From pni/pnicodeobj.h *****************/
  panSymbol_MAPFUNC (pniCodeobjName_read,        panCodeobjName_read)
  panSymbol_MAPFUNC (pniCodeobjName_write,       panCodeobjName_write)
  panSymbol_MAPFUNC (pniCodeobj_getLength,       panCodeobj_getLength)
  panSymbol_MAKEFUNC(pniSwseg_makePersistent)
  panSymbol_MAKEFUNC(pniSwseg_releasePersistent)
  panSymbol_MAKEFUNC(pniCodeobj_lookup)

  /***************** From pni/network.h *****************/
  panSymbol_MAPFUNC (pniNodeAddr_read, panNodeAddr_read)
  panSymbol_MAPFUNC (pniNodeAddr_write,  panNodeAddr_write)
  panSymbol_MAPFUNC (pniNodeAddr_compare,        panNodeAddr_compare)
  panSymbol_MAPFUNC (pniNodeAddr_stringify,      panNodeAddr_stringify)
  panSymbol_MAPFUNC (pniNodeAddr_getLocalAddr,   panNodeAddr_getLocalAddr)


  /***************** From pni/swseg.h *****************/
  panSymbol_MAKEFUNC(pniSwseg_create_byMalloc)
  panSymbol_MAKEFUNC(pniSwseg_overlay)
  panSymbol_MAPFUNC (pniSwseg_chain,              panSwseg_chain)
  panSymbol_MAPFUNC (pniSwseg_getSize,            panSwseg_getSize)
  panSymbol_MAPFUNC (pniSwseg_getChainLength,     panSwseg_getChainLength)
  panSymbol_MAPFUNC (pniSwseg_getBytesInChain,    panSwseg_getBytesInChain)
  panSymbol_MAPFUNC (pniSwseg_ref,                panSwseg_ref)
  panSymbol_MAPFUNC (pniSwseg_unref,              panSwseg_unref)

  /***************** From pni/pniutil_hashtab.h *****************/
  panSymbol_MAKEFUNC(pniHashTab_create)
  panSymbol_MAKEFUNC(pniHashTab_lookup)
  panSymbol_MAKEFUNC(pniHashTab_remove)
  panSymbol_MAKEFUNC(pniHashTab_insert)

  /***************** From pni/pnitimestamp.h *****************/
  panSymbol_MAPFUNC(pniTimestamp_read,           panTimestamp_read)
  panSymbol_MAPFUNC(pniTimestamp_write,          panTimestamp_write)
  panSymbol_MAPFUNC(pniTimestamp_stringify,      panTimestamp_stringify)
  panSymbol_MAPFUNC(pniTimestamp_compare,        panTimestamp_compare)
  panSymbol_MAPFUNC(pniTimestamp_subtract,       panTimestamp_subtract)
  panSymbol_MAPFUNC(pniTimestamp_normalize,      panTimestamp_normalize)
  panSymbol_MAPFUNC(pniTimestamp_getCurrent,     panTimestamp_getCurrent)
  panSymbol_MAPFUNC(pniTimestamp_getResolution,  panTimestamp_getResolution)

  /***************** From pni/swsstream.h *****************/
  panSymbol_MAPFUNC (pniStream_init,              panStream_init)
  panSymbol_MAPFUNC (pniStream_seekFwd,           panStream_seekFwd)
  panSymbol_MAPFUNC (pniStream_getBytesRemaining, panStream_getBytesRemaining)
  panSymbol_MAPFUNC (pniStream_getState,          panStream_getState)
  panSymbol_MAPFUNC (pniStream_copy,              panStream_copy)
  panSymbol_MAPFUNC (pniStream_read,              panStream_read)
  panSymbol_MAPFUNC (pniStream_write,             panStream_write)
  /* panSymbol_MAKEFUNC (pniStream_dup)  -- handled as a macro */
  panSymbol_MAPFUNC (pniStream_eofP,              panStream_eofP)

  panSymbol_MAPFUNC (pniStream_read_uint8,        panStream_read_uint8)
  panSymbol_MAPFUNC (pniStream_read_uint16,       panStream_read_uint16)
  panSymbol_MAPFUNC (pniStream_read_uint32,       panStream_read_uint32)
  panSymbol_MAPFUNC (pniStream_read_int8,         panStream_read_int8)
  panSymbol_MAPFUNC (pniStream_read_int16,        panStream_read_int16)
  panSymbol_MAPFUNC (pniStream_read_int32,        panStream_read_int32)

  panSymbol_MAPFUNC (pniStream_read_lsbfirst_uint8,  
		     panStream_read_uint8)
  panSymbol_MAPFUNC (pniStream_read_lsbfirst_uint16, 
		     panStream_read_lsbfirst_uint16)
  panSymbol_MAPFUNC (pniStream_read_lsbfirst_uint32, 
		     panStream_read_lsbfirst_uint32)
  panSymbol_MAPFUNC (pniStream_read_lsbfirst_int8,   
		     panStream_read_int8)
  panSymbol_MAPFUNC (pniStream_read_lsbfirst_int16,  
		     panStream_read_lsbfirst_int16)
  panSymbol_MAPFUNC (pniStream_read_lsbfirst_int32,  
		     panStream_read_lsbfirst_int32)

  panSymbol_MAPFUNC (pniStream_read_msbfirst_uint8,  
		     panStream_read_uint8)
  panSymbol_MAPFUNC (pniStream_read_msbfirst_uint16, 
		     panStream_read_msbfirst_uint16)
  panSymbol_MAPFUNC (pniStream_read_msbfirst_uint32, 
		     panStream_read_msbfirst_uint32)
  panSymbol_MAPFUNC (pniStream_read_msbfirst_int8,   
		     panStream_read_int8)
  panSymbol_MAPFUNC (pniStream_read_msbfirst_int16,  
		     panStream_read_msbfirst_int16)
  panSymbol_MAPFUNC (pniStream_read_msbfirst_int32,  
		     panStream_read_msbfirst_int32)
  
  panSymbol_MAPFUNC (pniStream_read_net_uint8,  panStream_read_uint8)
  panSymbol_MAPFUNC (pniStream_read_net_uint16, panStream_read_msbfirst_uint16)
  panSymbol_MAPFUNC (pniStream_read_net_uint32, panStream_read_msbfirst_uint32)
  panSymbol_MAPFUNC (pniStream_read_net_int8,   panStream_read_int8)
  panSymbol_MAPFUNC (pniStream_read_net_int16,  panStream_read_msbfirst_int16)
  panSymbol_MAPFUNC (pniStream_read_net_int32,  panStream_read_msbfirst_int32)

  panSymbol_MAPFUNC (pniStream_write_uint8,        panStream_write_uint8)
  panSymbol_MAPFUNC (pniStream_write_uint16,       panStream_write_uint16)
  panSymbol_MAPFUNC (pniStream_write_uint32,       panStream_write_uint32)
  panSymbol_MAPFUNC (pniStream_write_int8,         panStream_write_int8)
  panSymbol_MAPFUNC (pniStream_write_int16,        panStream_write_int16)
  panSymbol_MAPFUNC (pniStream_write_int32,        panStream_write_int32)

  panSymbol_MAPFUNC (pniStream_write_lsbfirst_uint8,  
		     panStream_write_uint8)
  panSymbol_MAPFUNC (pniStream_write_lsbfirst_uint16, 
		     panStream_write_lsbfirst_uint16)
  panSymbol_MAPFUNC (pniStream_write_lsbfirst_uint32, 
		     panStream_write_lsbfirst_uint32)
  panSymbol_MAPFUNC (pniStream_write_lsbfirst_int8,   
		     panStream_write_int8)
  panSymbol_MAPFUNC (pniStream_write_lsbfirst_int16,  
		     panStream_write_lsbfirst_int16)
  panSymbol_MAPFUNC (pniStream_write_lsbfirst_int32,  
		     panStream_write_lsbfirst_int32)

  panSymbol_MAPFUNC (pniStream_write_msbfirst_uint8,  
		     panStream_write_uint8)
  panSymbol_MAPFUNC (pniStream_write_msbfirst_uint16, 
		     panStream_write_msbfirst_uint16)
  panSymbol_MAPFUNC (pniStream_write_msbfirst_uint32, 
		     panStream_write_msbfirst_uint32)
  panSymbol_MAPFUNC (pniStream_write_msbfirst_int8,   
		     panStream_write_int8)
  panSymbol_MAPFUNC (pniStream_write_msbfirst_int16,  
		     panStream_write_msbfirst_int16)
  panSymbol_MAPFUNC (pniStream_write_msbfirst_int32,  
		     panStream_write_msbfirst_int32)

  panSymbol_MAPFUNC (pniStream_write_net_uint8,  panStream_write_uint8)
  panSymbol_MAPFUNC (pniStream_write_net_uint16, 
		     panStream_write_msbfirst_uint16)
  panSymbol_MAPFUNC (pniStream_write_net_uint32, 
		     panStream_write_msbfirst_uint32)
  panSymbol_MAPFUNC (pniStream_write_net_int8,   panStream_write_int8)
  panSymbol_MAPFUNC (pniStream_write_net_int16, 
		     panStream_write_msbfirst_int16)
  panSymbol_MAPFUNC (pniStream_write_net_int32,  
		     panStream_write_msbfirst_int32)

  




  { NULL, NULL }
};
