/* $Header: /home/dl2/nygren/cvs/pan/src/node/pni/pniapplink.c,v 1.3 1997/09/09 19:31:23 nygren Exp $
 *
 * FILE:    pniappif.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Implements routines that may be called from
 * capsules through the PAN Node Interface for sending
 * data to application links.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>

#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pniapplink.h>


/* Checks to see if the executing capsule was inserted into the node
 * by an application.  If it was, returns a pniAppLink that can be
 * used to send data to the applications.  Returns NULL if the capsule
 * wasn't insterted into the node by an application.  */
pniAppLink *
pniAppLink_getSender(void)
{
  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniAppLink_getSender: "
		     "called with no running capsule!\n");
      return NULL;
    }
 
  if (!panNode_CurrentCapsule->applink)
    return NULL;
 
  /* Since the applink is already in the capsules swseg ring,
   * we don't need to put it there again. */
  return (pniAppLink*)panNode_CurrentCapsule->applink;
}


