/* $Header: /home/dl2/nygren/cvs/pan/src/node/pni/pnicodeobj.c,v 1.3 1997/12/18 02:33:35 nygren Exp $
 *
 * FILE:    pnicodeobj.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Implements routines that may be called from
 * capsules through the PAN Node Interface for loading 
 * code objects into the node's code cache.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>

#include <pni/pnicodeobj.h>

/* TODO: Clean this up.  This is just gross... */
#ifdef __KERNEL__
extern int
pniCodeobj_load_enqueue(pniStream *startps, pniNodeAddr *homeaddr);
#endif /* __KERNEL__ */


/* Loads the code object starting at startps into the code cache.
 * Also takes the address of the node to go to to request dependencies from.
 * Should we advance startps???
 * Returns 1 on success, 0 on failure.
 *
 * Within the kernel, an alternate version is used which queues
 * up the codeobj to be loaded by a kernel thread.
 *
 */
int
pniCodeobj_load_immediate(pniStream *startps, pniNodeAddr *homeaddr)
{
  panCodeobj *co;
  
  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniCodeobj_load: "
		     "called with no running capsule!\n");
      return 0;
    }

#if 0
#ifdef __KERNEL__
  /* If we have an applink, there's no need to use the kpand
   * since we're not in an interrupt anyways.
   * (TODO: except it needs more locking if code loading routines can
   * also be entered here...)
   */
  if (!panNode_CurrentCapsule->applink) {
    return pniCodeobj_load_enqueue(startps, homeaddr);
  }
#endif  
#endif

  panPrint_debug("pniCodeobj_load_immediate: called from <%s> with homeaddr ",
		 panNode_CurrentCapsule->codeobj->symname);
  panNodeAddr_print(homeaddr);
  panPrint_debug("\n");

  co = panCodeobj_load(startps, panNode_CodeCache, 
		       homeaddr);
  
  if (co) return 1;
  else    return 0;
}



/* Looks a code object up in the code cache by name.  If it exists,
 * returns a software segment containing the code object.  Otherwise,
 * returns null.  Stores the length of the code object into the value
 * pointed to by colength if colength!=NULL.  Returns NULL if
 * the code object isn't in the code cache.
 *
 *   *** IMPORTANT *** The swseg returned is the same one that
 *   contains the executable stuff.  It is critical that it
 *   not get modified.  This MUST be read-only in any secure
 *   implementation.
 */
pniSwseg *
pniCodeobj_lookup(pniCodeobjName *coname, int *colength)
{
  panCodeobj *co;
  int         len;

  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniCodeobj_lookup: "
		     "called with no running capsule!\n");
      return NULL;
    }
  if (!coname) 
    {
      panPrint_warn("pniCodeobj_lookup: called with NULL coname\n");
      return NULL;
    }

  co = panCodeCache_lookup(panNode_CodeCache, coname);
  
  if (!co 
      || !co->codeobj
      || (!panCodeobj_STATUS_IS_READY_P(co->status)
	  && !panCodeobj_STATUS_NEED_DEPENDS_P(co->status)))
      
    { return NULL; }

  len = panSwseg_getBytesInChain(co->codeobj);
  if (colength) 
    *colength = len;
  
  panSwseg_ref(co->codeobj);
  panSwsegRing_add(panNode_CurrentCapsule->sring, co->codeobj);

  return co->codeobj;
}
