/* $Header: /home/dl2/nygren/cvs/pan/src/node/symtab.h,v 1.2 1997/07/15 22:00:37 nygren Exp $
 *
 * FILE:    symtab.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Routines for storing an looking up symbols.
 * A symbol entry has a string name and a pointer to 
 * a function or a variable.
 *
 */

#ifndef NODE_SYMTAB_H_INC
#define NODE_SYMTAB_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <node/nodecodeobj.h>


/* A symbol contains a string symbol name and a function
 * pointer to either a function or to a variable.  (Note that
 * according to the ANSI C spec, a function pointer may be 
 * larger than a normal pointer.
 */
typedef struct _panSymbol {
  char	   *name;		/* A null-terminated string naming a symbol */
  void	  (*addr)();		/* A pointer to the object that name names */
} panSymbol;


/* A symbol table entry contains a symbol and a pointer to the 
 * next entry in the table. */
typedef struct _panSymTabEntry {
  panSymbol               sym;	/* The symbol information */
  struct _panSymTabEntry *next;	/* The next entry in the symbol table */
} panSymTabEntry;

/* A symbol table contains a list of panSymTabEntry elements. */
typedef struct _panSymTab {
  panSymTabEntry *head;	/* The first entry in the symtab */
} panSymTab;


/* The symbol table of exported global symbols */
extern panSymTab *panSymTab_GlobalSymTab;

/* A list of symbols to initialize the global symbol table with */
extern panSymbol panSymTab_GlobalSymbols_PlatformIndependent[];
extern panSymbol panSymTab_GlobalSymbols_PlatformDependent[];

extern panSymTab *
panSymTab_create(void);

extern void
panSymTab_insert(panSymTab *symtab, panSymbol *sym);

extern panSymbol *
panSymTab_lookup(panSymTab *symtab, char *name);

extern void
panSymTab_destroy(panSymTab *symtab);

/*  PROTO in nodecodeobj.h
extern panSymbol *
panSymTab_lookupInContext(panCodeobj *context, char *name);
*/

/* Macros for creating lists of panSymbols */
#define panSymbol_MAPFUNC( name, func ) { #name, (void (*)())&func },
#define panSymbol_MAKEFUNC( func ) { #func, (void (*)())&func },
#define panSymbol_MAPVAR ( name, var )  { #name, (void (*)())&var },
#define panSymbol_MAKEVAR ( var )  { #var, (void (*)())&var },


#endif /* NODE_SYMTAB_H_INC */

