/* $Header: /home/dl2/nygren/cvs/pan/src/node/test/testloader.c,v 1.3 1997/12/10 06:58:43 nygren Exp $
 *
 * FILE:    testloader.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Tests the code loader routines for PAN.
 * This was quickly hacked together as a test driver.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>

/* If the predicate is true, prints the error and exits */
#define REQUIRE_NOT_PRED(pred, errmsg) \
     do {if (pred) { panPrint_error("%s:%d: %s\n",__FILE__,__LINE__,errmsg); \
	exit(-1); } } while(0);

#define ADDTOKEYRING(x) 

panCodeobj *load_codeobj(char *infilename, panCodeCache *cache);

void
main(int argc, char *argv[])
{
  panCodeCache *codecache;
  int i;

  codecache = panCodeCache_create();
  REQUIRE_NOT_PRED(!codecache, "creation of codecache failed!");

  for (i=1; i<argc; i++)
    load_codeobj(argv[i], codecache);
  
}

panCodeobj *
load_codeobj(char *infilename, panCodeCache *cache)
{
  int       infd;		/* Input file descriptor */
  struct stat infstatbuf;	/* Stat info about input file */
  unsigned totalsize;
  panSwseg  *cosws;		/* Software segment for code object */
  panCodeobj *co;
  panStream   _ps, *ps=&_ps;
  panSwsegRing *swsring;

  panPrint_msg("\n\n********* load_codeobj %s: ***********\n", infilename);

  /* Get the size of the input file */
  if (0 != stat(infilename, &infstatbuf))
    { perror("stat");
      panPrint_error("\t*** couldn't stat input file %s\n", 
		     infilename);  return(0);  }
  totalsize = infstatbuf.st_size;
  if (ADDR_NOT_ALIGNED_P(totalsize))
    { panPrint_error("\t*** invalid code object (size of is not 4-byte word aligned).\n"); return(0); }

  /* Now open and create a swseg by mmapping in the file */
  infd = open(infilename, O_RDONLY);
  if (-1 == infd)
    { perror("open"); panPrint_error("\t*** couldn't open file for reading\n"); return(0); }

  panPrint_msg("\thas size %d\n", totalsize);

  swsring = panSwsegRing_create(8);
  REQUIRE_NOT_PRED(!swsring, "creation of swsring failed!");

  cosws = panSwseg_create_byMmap(totalsize,
				  PROT_READ, MAP_PRIVATE, infd, 0);
  REQUIRE_NOT_PRED(!cosws, "creation of binseg failed!");
  panSwsegRing_add(swsring, cosws);
  close(infd);			/* doesn't need to stay open to 
				 * keep the mapping */
  
  
  panStream_init(ps, cosws, 0);
  co = panCodeobj_load(ps, cache, NULL);
  if (!co)
    panPrint_warn("*** load_codeobj: load of codeobj failed!\n");

  panPrint_msg("Done with %s\n", infilename);
  panSwsegRing_destroy(swsring);

  return(co);
}
