/* $Header: /home/dl2/nygren/cvs/pan/src/panotool/misc.c,v 1.2 1997/06/01 00:52:52 nygren Exp $
 *
 * FILE:    misc.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Misc utility routines used by panotool.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>

#include "misc.h"

/* Rounds an address up to the nearest aligned address */
unsigned long ADDR_ROUND_UP(unsigned long x)
{
  return (((x)%4)?(x+(4-(x%4))):x);
}     


/* Creates a new and empty stringlist */
stringlist *
stringlist_create(void)
{
  stringlist *sl = malloc(sizeof(stringlist));

  if (sl == NULL) 
    { fprintf(stderr, "Couldn't allocate stringlist\n"); return NULL; }
  sl->head = sl->tail = NULL;
  return sl;
}

/* Frees memory associated with a stringlist */
void
stringlist_destroy(stringlist *sl)
{
  stringlistel *e1, *e2;

  if (!sl) return;
  for (e1 = sl->head; e1; e1 = e2)
    {
      e2 = e1->next;
      free(e1);
    }
  free(sl);
}

/* Adds the string addstr to the stringlist sl.
 * If dup is true, adds a duplicate of the string rather than the
 * string itself. */
void        
stringlist_addtotail(stringlist *sl, char *addstr, int dup)
{
  stringlistel *el = malloc(sizeof(stringlistel));

  if (el == NULL) 
    { fprintf(stderr, "Couldn't allocate stringlist\n");   return; }
  if (dup && addstr)
    {
      char *newstr = malloc(strlen(addstr)+1);
      strcpy(newstr, addstr);
      addstr = newstr;
    }
  el->next = NULL;
  el->str = addstr;
  if (!sl->tail)
    { 
      sl->head = sl->tail = el;
    }
  else
    {
      sl->tail->next = el;
      sl->tail = el;
    }
}

/* Returns the head of a stringlist.
 * The "next" fields can be used for iterating. */
stringlistel *
stringlist_gethead(stringlist *sl)
{
  return sl->head;
}

/* Returns 1 if base ends in ext.  Else returns 0. */
int
string_extension_compare(const char *s, const char *ext)
{
  int slen, extlen;

  if (!s || !ext) return 0;
  slen = strlen(s);
  extlen = strlen(ext);
  if (slen < extlen) return 0;
  return (0 == strcmp(ext, s+(slen-extlen)));
}


/* Takes a filename and seeks past any path components and
 * returns a pointer to that offset into fn.  */
char *
string_file_basename(char *fn)
{
  char *pos;  /* The position in fn past the last "/" */

  if (!fn) return NULL;
  for (pos=fn; *fn; fn++)
    {
      if (*fn == '/') 
	pos = fn + 1;
    }
  return pos;
}

