#include "common/sysconfig.h"
#include "common/sha1.h"

/* SHA1 test vectors, from Handbook of Applied Cryptography Menezes,
   Oorschot and Vanstone */

#define NTEST 6
struct {
   char* in;                    /* input */
   u_int8_t res[SHA1_SIZE];     /* expected result */
}tv[NTEST] = {
   {"", {0xda, 0x39, 0xa3, 0xee, 0x5e, 
         0x6b, 0x4b, 0x0d, 0x32, 0x55, 
         0xbf, 0xef, 0x95, 0x60, 0x18, 
         0x90, 0xaf, 0xd8, 0x07, 0x09}},
   {"a",{0x86, 0xf7, 0xe4, 0x37, 0xfa, 
         0xa5, 0xa7, 0xfc, 0xe1, 0x5d, 
         0x1d, 0xdc, 0xb9, 0xea, 0xea, 
         0xea, 0x37, 0x76, 0x67, 0xb8}},
   {"abc",{0xA9, 0x99, 0x3e, 0x36, 0x47,
           0x06, 0x81, 0x6a, 0xba, 0x3e,
           0x25, 0x71, 0x78, 0x50, 0xc2,
           0x6c, 0x9c, 0xd0, 0xd8, 0x9d}},
   {"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
    {0x84, 0x98, 0x3E, 0x44,
     0x1C, 0x3B, 0xD2, 0x6E,
     0xBA, 0xAE, 0x4A, 0xA1,
     0xF9, 0x51, 0x29, 0xE5,
     0xE5, 0x46, 0x70, 0xF1}},
   {"abcdefghijklmnopqrstuvwxyz", { 0x32, 0xd1, 0x0c, 0x7b, 0x8c, 
                                    0xf9, 0x65, 0x70, 0xca, 0x04, 
                                    0xce, 0x37, 0xf2, 0xa1, 0x9d, 
                                    0x84, 0x24, 0x0d, 0x3a, 0x89}},
   /* a million a's, 10**5 of this string */
   {"aaaaaaaaaa", {0x34, 0xAA, 0x97, 0x3C,
                   0xD4, 0xC4, 0xDA, 0xA4,
                   0xF6, 0x1E, 0xEB, 0x2B,
                   0xDB, 0xAD, 0x27, 0x31,
                   0x65, 0x34, 0x01, 0x6F}},
};

void printbs( u_char* bs, int len)
{
   u_char* end = bs + len;
   while(bs < end) printf("0x%02x, ", *bs++);
   printf("\n");
}
int main()
{
   sha1_ctx c;
   u_int8_t h[SHA1_SIZE];
   char buf[100];
   int i,j;

   for(i = 0; i < NTEST-1; i++) {
      sha1_init(&c);
      strncpy(buf, tv[i].in, 100);
      sha1_update(&c, (u_char*)buf, strlen(buf));
      sha1_final(h, &c);
      if( strcmp(tv[i].in, buf) ) { 
         printf("%s--is now--%s\n", tv[i].in, buf); 
         exit(1);
      }
      if( memcmp(h, tv[i].res, SHA1_SIZE) ) { 
         printf("h(%s) = ", buf); printbs(h, SHA1_SIZE);
         exit(1);
      }
   }
   strncpy(buf, tv[i].in, 100);
   sha1_init(&c);
   for( j = 0; j < 100000; j++ ){
      sha1_update(&c,  (u_char*)buf, strlen(buf));
   }
   sha1_final(h,&c);
   if( strcmp(tv[i].in, buf) ) { 
      printf("%s--is now--%s\n", tv[i].in, buf); 
      exit(1);
   }
   if( memcmp(h, tv[i].res, SHA1_SIZE) ) { 
      printf("h(%s) = ", buf); printbs(h, SHA1_SIZE);
      exit(1);
   }
   printf("passed\n");
   return 0;
}
