/* $Header: /home/dl2/nygren/cvs/pan/src/test/timestamp/timestamp_test.c,v 1.1 1998/01/18 00:12:05 nygren Exp $
 *
 * FILE:    timestamp_test.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for testing timestamp data abstractions
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/timestamp.h>

#define TEST_PRED(pred,printstatement) \
   do { if ((pred)) { printf("%s:%d: ", __FILE__, __LINE__); \
                      printstatement; \
		      exit(-1); } } while(0)		


char tsstrbuf[panTimestamp_STRINGIFY_BUFSIZE];
#define ts2str(ts,fmt) (panTimestamp_stringify(tsstrbuf, \
				     panTimestamp_STRINGIFY_BUFSIZE, \
				     (ts), (fmt)))

void
test_print_ts(panTimestamp *ts);

void
perftest(void);

void
main(int argc, char **argv)
{
  panTimestamp a, b, c, d;
  

#undef TEST_PRINT
#ifdef TEST_PRINT
  a.sec = 56; a.usec = 1345;
  test_print_ts(&a);
  a.sec = 56; a.usec = 0;
  test_print_ts(&a);
  a.sec = 0; a.usec = 1345;
  test_print_ts(&a);
  a.sec = 5; a.usec = -1;
  test_print_ts(&a);
  a.sec = -5; a.usec = 4357;
  test_print_ts(&a);
  a.sec = 5; a.usec = -3000001;
  test_print_ts(&a);
  a.sec = 5; a.usec =  3000135;
  test_print_ts(&a);
#endif /* TEST_PRINT */

  a.sec = 56; a.usec = 1345;  panTimestamp_dup(&b,&a);
  panTimestamp_normalize(&a);
  TEST_PRED(a.sec!=56||a.usec!=1345, 
	   printf("panTimestamp_normalize: failed on <%d,%d>\n",b.sec,b.usec));
  a.sec = 56; a.usec = 0;  panTimestamp_dup(&b,&a);
  panTimestamp_normalize(&a);
  TEST_PRED(a.sec!=56||a.usec!=0, 
	   printf("panTimestamp_normalize: failed on <%d,%d>\n",b.sec,b.usec));
  a.sec = 0; a.usec = 1345;  panTimestamp_dup(&b,&a);
  panTimestamp_normalize(&a);
  TEST_PRED(a.sec!=0||a.usec!=1345, 
	   printf("panTimestamp_normalize: failed on <%d,%d>\n",b.sec,b.usec));
  a.sec = 5; a.usec = -2;  panTimestamp_dup(&b,&a);
  panTimestamp_normalize(&a);
  TEST_PRED(a.sec!=4||a.usec!=999998, 
	   printf("panTimestamp_normalize: failed on <%d,%d>\n",b.sec,b.usec));
  a.sec = -5; a.usec = 4357;  panTimestamp_dup(&b,&a);
  panTimestamp_normalize(&a);
  TEST_PRED(a.sec!=-5||a.usec!=4357, 
	   printf("panTimestamp_normalize: failed on <%d,%d>\n",b.sec,b.usec));
  a.sec = 5; a.usec = -3000001;  panTimestamp_dup(&b,&a);
  panTimestamp_normalize(&a);
  TEST_PRED(a.sec!=1||a.usec!=999999, 
	   printf("panTimestamp_normalize: failed on <%d,%d>\n",b.sec,b.usec));
  a.sec = 5; a.usec =  3000135;
  panTimestamp_normalize(&a);  panTimestamp_dup(&b,&a);
  TEST_PRED(a.sec!=8||a.usec!=135, 
	   printf("panTimestamp_normalize: failed on <%d,%d>\n",b.sec,b.usec));
  
  a.sec=5; a.usec=56;  b.sec=3; b.usec=56;
  TEST_PRED(1!=panTimestamp_compare(&a,&b), 
	    printf("panTimestamp_compare: failed\n"));
  a.sec=5; a.usec=56;  b.sec=5; b.usec=56;
  TEST_PRED(0!=panTimestamp_compare(&a,&b), 
	    printf("panTimestamp_compare: failed\n"));
  a.sec=3; a.usec=56;  b.sec=5; b.usec=56;
  TEST_PRED(-1!=panTimestamp_compare(&a,&b), 
	    printf("panTimestamp_compare: failed\n"));
  a.sec=3; a.usec=25;  b.sec=3; b.usec=56;
  TEST_PRED(-1!=panTimestamp_compare(&a,&b), 
	    printf("panTimestamp_compare: failed\n"));
  a.sec=3; a.usec=56;  b.sec=3; b.usec=25;
  TEST_PRED(1!=panTimestamp_compare(&a,&b), 
	    printf("panTimestamp_compare: failed\n"));

  a.sec=3; a.usec=6;  b.sec=5; b.usec=2;
  panTimestamp_subtract(&c,&a,&b);
  TEST_PRED(c.sec!=-2||c.usec!=4,
	    printf("panTimestamp_subtract: failed with %d,%d\n",c.sec,c.usec));
  a.sec=5; a.usec=10;  b.sec=2; b.usec=2;
  panTimestamp_subtract(&c,&a,&b);
  TEST_PRED(c.sec!=3||c.usec!=8,
	    printf("panTimestamp_subtract: failed with %d,%d\n",c.sec,c.usec));
  a.sec=5; a.usec=2;  b.sec=2; b.usec=10;
  panTimestamp_subtract(&c,&a,&b);
  TEST_PRED(c.sec!=2||c.usec!=999992,
	    printf("panTimestamp_subtract: failed with %d,%d\n",c.sec,c.usec));

  perftest();

  printf("timestamp_test: passed!\n");
  exit(0);
}



void
test_print_ts(panTimestamp *ts)
{
  panPrint_msg("For ts<%d,%d> printing <%s>  (secs,pad,pad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INSECS
		      |panTimestamp_STRINGIFY_FMT_PADSECS
		      |panTimestamp_STRINGIFY_FMT_PADUSECS));
  panPrint_msg("For ts<%d,%d> printing <%s>  (usecs,pad,pad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INUSECS
		      |panTimestamp_STRINGIFY_FMT_PADSECS
		      |panTimestamp_STRINGIFY_FMT_PADUSECS));
  panPrint_msg("For ts<%d,%d> printing <%s>  (secs,nopad,pad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INSECS
		      |panTimestamp_STRINGIFY_FMT_PADUSECS));
  panPrint_msg("For ts<%d,%d> printing <%s>  (usecs,nopad,pad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INUSECS
		      |panTimestamp_STRINGIFY_FMT_PADUSECS));

  panPrint_msg("For ts<%d,%d> printing <%s>  (secs,pad,nopad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INSECS
		      |panTimestamp_STRINGIFY_FMT_PADSECS));
  panPrint_msg("For ts<%d,%d> printing <%s>  (usecs,pad,nopad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INUSECS
		      |panTimestamp_STRINGIFY_FMT_PADSECS));
  panPrint_msg("For ts<%d,%d> printing <%s>  (secs,nopad,nopad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INSECS));
  panPrint_msg("For ts<%d,%d> printing <%s>  (usecs,nopad,nopad)\n", 
	       ts->sec, ts->usec, 
	       ts2str(ts,panTimestamp_STRINGIFY_FMT_INUSECS));
  panPrint_msg("--------\n");
}


/* Tries to approximate the resolution of the time
 * routines as well as determining how long they take to run.
 * Runs for three seconds. */
void
perftest(void)
{
  panTimestamp start, last, cur, maxruntime, maxdiff, diff, runtime, mindiff;
  unsigned int itercount=0, diffaccum=0;
  
  maxdiff.sec=0; maxdiff.usec=0;
  mindiff.sec=100000; mindiff.usec=0;
  maxruntime.sec=2; maxruntime.usec=0;
  printf("timestamp_test: running performance tests.  should take %d secs.\n",
	 maxruntime.sec);
  panTimestamp_getCurrent(&start);
  panTimestamp_dup(&last, &start);  
  do {
    panTimestamp_getCurrent(&cur);
    panTimestamp_subtract(&diff, &cur, &last);
    diffaccum += diff.usec + diff.sec*1000000;
    if (panTimestamp_compare(&diff, &maxdiff) > 0)
      {
	panTimestamp_dup(&maxdiff, &diff);
      }
    if (panTimestamp_compare(&diff, &mindiff) < 0)
      {	
	panTimestamp_dup(&mindiff, &diff);
      }
    panTimestamp_subtract(&runtime, &cur, &start);    
    panTimestamp_dup(&last, &cur);  
    itercount++;
  } while (panTimestamp_compare(&maxruntime, &runtime) > 0);

  printf("timestamp_test: iterations=%d  maxdiff=%susecs\n",
	 itercount, ts2str(&maxdiff,0));
  printf("timestamp_test:   mindiff=%susecs  meanruntime=%dusecs\n",
	 ts2str(&mindiff,0), diffaccum/itercount);
  if (diffaccum/itercount > 500)
    printf("timestamp_test.c:%d: warning: gettimeofday provided by OS may not have near-microsecond resolution\n", __LINE__);
  printf("timestamp_test: done with performance tests.\n");
}
